"""A Singleton Plugin Manager Class: Manages the loading and retrieval of Workbench plugins"""

import os
import sys
import atexit
import shutil
import tempfile
import logging
import importlib
import threading
from typing import Union, Dict, List, Any, Optional
from types import ModuleType

# Workbench Imports
from workbench.utils.config_manager import ConfigManager
from workbench.utils.s3_utils import copy_s3_files_to_local
from workbench.web_interface.page_views.page_view import PageView
from workbench.web_interface.components.plugin_interface import PluginInterface
from workbench.web_interface.components.plugin_interface import PluginPage


class PluginManager:
    """A Singleton Plugin Manager Class: Manages the loading and retrieval of Workbench plugins"""

    _instance = None

    def __new__(cls, *args, **kwargs):
        """Create a new instance of the class if it doesn't exist, else return the existing instance."""
        if cls._instance is None:
            cls._instance = super(PluginManager, cls).__new__(cls)
            cls._instance.__initialized = False  # Gets initialized in __init__
        return cls._instance

    def __init__(self):
        """Initialize the PluginManager as a singleton."""
        if self.__initialized:
            return

        self.log = logging.getLogger("workbench")
        self.config_plugin_dir = None
        self.loading_dir = False
        self.plugin_modified_time = None
        self.plugins: Dict[str, dict] = {"components": {}, "transforms": {}, "views": {}, "pages": {}, "css": {}}

        # Get the plugin directory from the config
        cm = ConfigManager()
        self.config_plugin_dir = cm.get_config("WORKBENCH_PLUGINS")
        if not self.config_plugin_dir:
            self.log.warning("WORKBENCH_PLUGINS not set. No plugins will be loaded.")
            return

        # Load the plugins
        self.load_plugins()

        # Singleton is now initialized
        self.__initialized = True

    def load_plugins(self):
        """Loads plugins from our 'load_from' directory"""

        # Check if the config directory is local file or S3
        if not self.config_plugin_dir.startswith("s3://"):
            self.loading_dir = self.config_plugin_dir

        # Load plugins from S3 (copy to a temporary directory, then load)
        else:
            self.loading_dir = tempfile.mkdtemp()
            copy_s3_files_to_local(self.config_plugin_dir, self.loading_dir)

            # Cleanup the temporary directory on exit
            atexit.register(self._cleanup_temp_dir)

        # Add the loading directory to the PYTHONPATH for custom packages
        sys.path.append(os.path.join(self.loading_dir, "packages"))

        self.log.important(f"Loading plugins from {self.loading_dir}...")
        for plugin_type in self.plugins.keys():
            self._load_plugins(self.loading_dir, plugin_type)

        # Store the most recent modified time
        self.plugin_modified_time = self._most_recent_modified_time()

    def _load_plugins(self, base_dir: str, plugin_type: str):
        """Internal: Load plugins of a specific type from a subdirectory.

        Args:
            base_dir (str): Base directory of the plugins.
            plugin_type (str): Type of the plugin to load (e.g., components, pages, transforms).
        """
        plugin_dir = os.path.join(base_dir, plugin_type)
        if not os.path.isdir(plugin_dir):
            return

        # For every file in the plugin directory
        for filename in os.listdir(plugin_dir):

            # Normal plugin loading
            module = self._load_module(plugin_dir, filename)
            if module is None:
                self.log.warning(f"Failed to load plugin: '{filename}' skipping...")
                continue

            # Now we have a module, let's iterate through its attributes
            for attr_name in dir(module):
                attr = getattr(module, attr_name)

                # Check if the attribute is a class and is defined in the module
                if isinstance(attr, type) and attr.__module__ == module.__name__:
                    self.log.important(f"Loading {plugin_type} plugin: {attr_name}")

                    # For web components, check if the class is a subclass of PluginInterface
                    if plugin_type == "components":
                        if issubclass(attr, PluginInterface):
                            self.plugins[plugin_type][attr_name] = attr
                        else:
                            # PluginInterface has additional information for failed validation
                            valid, validation_error = PluginInterface.validate_subclass(attr)
                            self.log.error(f"Plugin '{attr_name}' failed validation:")
                            self.log.error(f"\tFile: {os.path.join(plugin_dir, filename)}")
                            self.log.error(f"\tClass: {attr_name}")
                            self.log.error(f"\tDetails: {filename} {validation_error}")

                    # For views, check if the class is a subclass of PageView
                    elif plugin_type == "views" and issubclass(attr, PageView):
                        self.plugins[plugin_type][attr_name] = attr

                    # For pages, check if the class has the required page plugin method (page_setup)
                    elif plugin_type == "pages":
                        if hasattr(attr, "page_setup"):
                            self.plugins[plugin_type][attr_name] = attr
                        else:
                            self.log.warning(f"Class {attr_name} in {filename} does not have all required page methods")

                    # Unexpected type
                    else:
                        self.log.error(f"Unexpected plugin type '{plugin_type}' for plugin '{attr_name}'")

    def _load_module(self, dir_path: str, filename: str) -> Optional[ModuleType]:
        """Internal: Load a module from a file"""
        try:
            if filename.endswith(".py") and not filename.startswith("_"):
                file_path = os.path.join(dir_path, filename)
                module_name = filename[:-3]
                spec = importlib.util.spec_from_file_location(module_name, file_path)
                if spec and spec.loader:
                    module = importlib.util.module_from_spec(spec)
                    spec.loader.exec_module(module)
                    return module
        except Exception as e:
            # Log or handle the exception as needed
            self.log.critical(f"Failed to load plugin: '{filename}': {e}")
        return None

    def get_all_plugins(self) -> Dict[str, dict[Any]]:
        """
        Retrieve a dictionary of all plugins.

        Returns:
            Dict[str, dict]: A dictionary of all plugins.
                {'pages': {name: plugin, name2: plugin2,...
                 'views': {name: plugin, name2: plugin2,...},
        """
        return self.plugins

    def get_list_of_web_plugins(self, plugin_page: PluginPage = None) -> List[Any]:
        """
        Retrieve a list of plugins for a specific page (e.g., MODEL, ENDPOINT).

        Args:
            plugin_page (PluginPage): Get plugins for a specific page (e.g., MODEL, ENDPOINT).

        Returns:
            List[Any]: A list of INSTANTIATED plugin classes for the requested page.
        """
        plugin_classes = [
            self.plugins["components"][x]
            for x in self.plugins["components"]
            if self.plugins["components"][x].auto_load_page == plugin_page
        ]
        return [x() for x in plugin_classes]

    def get_web_plugin(self, plugin_name: str) -> PluginInterface:
        """
        Retrieve a specific web plugin by name

        Args:
            plugin_name (str): The name of the web plugin to retrieve

        Returns:
            PluginInterface: The INSTANTIATED web plugin class with the given name
        """
        web_plugin = self.plugins["components"].get(plugin_name)
        return web_plugin() if web_plugin else None

    def get_view(self, view_name: str) -> Union[PageView, None]:
        """
        Retrieve a page view plugin with the given name.

        Args:
            view_name (str): The name of the page view to retrieve (None if not found).

        Returns:
            PageView: An INSTANTIATED view class with the given name.
        """
        view = self.plugins["views"].get(view_name)
        return view() if view else None

    def get_pages(self) -> dict:
        """
        Retrieve a dict of plugins pages

        Returns:
            dict: A dict of INSTANTIATED plugin pages.
        """
        pages = self.plugins["pages"]
        return {name: page() for name, page in pages.items()}

    def get_pages_with_timeout(self) -> dict:
        """
        Retrieve a dict of plugin pages with a timeout mechanism using threads.

        Returns:
            dict: A dict of instantiated plugin pages or excludes pages that take too long.
        """
        pages = self.plugins["pages"]  # Dictionary of page name to page class
        instantiated_pages = {}
        timeout_seconds = 10  # Timeout threshold

        def instantiate_page(name, page_func):
            """Attempt to instantiate a page and log errors if it fails or times out."""
            result = [None]
            thread = threading.Thread(target=lambda: result.__setitem__(0, page_func()))
            thread.start()
            thread.join(timeout_seconds)

            if thread.is_alive():
                self.log.critical(f"Page {name} load longer than {timeout_seconds} seconds...")
            else:
                instantiated_pages[name] = result[0]

        # Instantiate each page with the timeout mechanism
        for name, page in pages.items():
            try:
                instantiate_page(name, page)
            except Exception as e:
                logging.critical(f"Error instantiating page '{name}': {e}")

        return instantiated_pages

    def _cleanup_temp_dir(self):
        """Cleans up the temporary directory created for S3 files."""
        if self.loading_dir and os.path.isdir(self.loading_dir):
            self.log.important(f"Cleaning up temporary directory: {self.loading_dir}")
            shutil.rmtree(self.loading_dir)
            self.loading_dir = None

    def plugins_modified(self) -> bool:
        """Check if the plugins have been modified since the last check

        Returns:
            bool: True if the plugins have been modified, else False
        """
        most_recent_time = self._most_recent_modified_time()
        if most_recent_time > self.plugin_modified_time:
            self.log.important("Plugins have been modified")
            self.plugin_modified_time = most_recent_time
            return True
        return False

    # Helper method to compute the most recent modified time
    def _most_recent_modified_time(self) -> float:
        """Internal: Compute the most recent modified time of a directory"""
        most_recent_time = 0
        for root, _, files in os.walk(self.loading_dir):
            for file in files:
                file_path = os.path.join(root, file)
                file_modified_time = os.path.getmtime(file_path)
                if file_modified_time > most_recent_time:
                    most_recent_time = file_modified_time
        return most_recent_time

    def __repr__(self) -> str:
        """String representation of the PluginManager state and contents

        Returns:
            str: String representation of the PluginManager state and contents
        """
        summary = "WORKBENCH_PLUGINS: " + self.config_plugin_dir + "\n"
        summary += "Plugins:\n"
        for plugin_type, plugin_dict in self.plugins.items():
            for name, plugin in plugin_dict.items():
                summary += f"  {plugin_type}: {name}: {plugin}\n"
        return summary


if __name__ == "__main__":
    """Exercise the PluginManager class"""
    from pprint import pprint

    # Create the class, load plugins, and call various methods
    manager = PluginManager()

    # Get web components for the model page
    model_plugin = manager.get_list_of_web_plugins(PluginPage.MODEL)

    # Get web components for the endpoint page
    endpoint_plugin = manager.get_list_of_web_plugins(PluginPage.ENDPOINT)

    # Get view plugin
    my_view = manager.get_view("ModelPluginView")

    # Get plugin pages
    plugin_pages = manager.get_pages()

    # Get all the plugins
    pprint(manager.get_all_plugins())

    # Get a web plugin by name
    my_plugin = manager.get_web_plugin("CustomTurbo")
    print(my_plugin)

    # Test REPR
    print(manager)

    # Test Modified Time by modifying a plugin file
    cm = ConfigManager()
    plugin_path = cm.get_config("WORKBENCH_PLUGINS") + "/components/endpoint_turbo.py"
    print(manager.plugins_modified())
    # Modify the plugin file modified time to now
    os.utime(plugin_path, None)
    print(manager.plugins_modified())

    # Test S3 Plugin Loading
    print("\n\n*** Testing S3 Plugin Loading ***\n\n")
    s3_path = "s3://sandbox-workbench-artifacts/workbench_plugins"
    cm = ConfigManager()
    cm.set_config("WORKBENCH_PLUGINS", s3_path)

    # Since we're using a singleton, we need to create a new instance
    PluginManager._instance = None
    manager = PluginManager()

    # Get web components for the model view
    model_plugin = manager.get_list_of_web_plugins(PluginPage.MODEL)

    # Get web components for the endpoint view
    endpoint_plugin = manager.get_list_of_web_plugins(PluginPage.ENDPOINT)

    # Get view plugin
    my_view = manager.get_view("ModelPluginView")

    # Get plugin pages
    plugin_pages = manager.get_pages()

    # Get all the plugins
    pprint(manager.get_all_plugins())



"""Tests for the AWS Account Clamp"""

# Workbench Imports
from workbench.core.cloud_platform.aws.aws_account_clamp import AWSAccountClamp


def test():
    """Tests for the AWS Account Clamp"""

    # Create the class
    aws_account_clamp = AWSAccountClamp()

    # Check out that AWS Account Clamp is working AOK
    """Check if the AWS Account is Setup Correctly"""
    print("*** AWS Caller/Base Identity Check ***")
    aws_account_clamp.check_aws_identity()
    print("Caller/Base Identity Check Success...")

    print("*** AWS Assumed Role Check ***")
    aws_account_clamp.check_assumed_role()

    print("*** AWS Workbench Bucket Check ***")
    aws_account_clamp.check_workbench_bucket()
    print("Workbench Bucket Check Success...")

    print("*** AWS Sagemaker Session/Client Check ***")
    sm_client = aws_account_clamp.sagemaker_client()
    print(sm_client.list_feature_groups()["FeatureGroupSummaries"])



"""CSVToDataSource: Class to move local CSV Files into a Workbench DataSource"""

import os
import pandas as pd

# Local imports
from workbench.core.transforms.transform import Transform, TransformInput, TransformOutput
from workbench.core.transforms.pandas_transforms.pandas_to_data import PandasToData
from workbench.utils.pandas_utils import convert_object_columns


class CSVToDataSource(Transform):
    """CSVToDataSource: Class to move local CSV Files into a Workbench DataSource

    Common Usage:
        ```python
        csv_to_data = CSVToDataSource(csv_file_path, data_name)
        csv_to_data.set_output_tags(["abalone", "csv", "whatever"])
        csv_to_data.transform()
        ```
    """

    def __init__(self, csv_file_path: str, data_name: str):
        """CSVToDataSource: Class to move local CSV Files into a Workbench DataSource

        Args:
            csv_file_path (str): The path to the CSV file to be transformed
            data_name (str): The Name of the Workbench DataSource to be created
        """

        # Call superclass init
        super().__init__(csv_file_path, data_name)

        # Set up all my instance attributes
        self.input_type = TransformInput.LOCAL_FILE
        self.output_type = TransformOutput.DATA_SOURCE

    def transform_impl(self, overwrite: bool = True):
        """Convert the local CSV file into Parquet Format in the Workbench Data Sources Bucket, and
        store the information about the data to the AWS Data Catalog workbench database
        """

        # Report the transformation initiation
        csv_file = os.path.basename(self.input_name)
        self.log.info(f"Starting {csv_file} -->  DataSource: {self.output_name}...")

        # Read in the Local CSV as a Pandas DataFrame
        df = pd.read_csv(self.input_name, low_memory=False)
        df = convert_object_columns(df)

        # Use the Workbench Pandas to Data Source class
        pandas_to_data = PandasToData(self.output_name)
        pandas_to_data.set_input(df)
        pandas_to_data.set_output_tags(self.output_tags)
        pandas_to_data.add_output_meta(self.output_meta)
        pandas_to_data.transform()

        # Report the transformation results
        self.log.info(f"{csv_file} -->  DataSource: {self.output_name} Complete!")

    def post_transform(self, **kwargs):
        """Post-Transform"""
        self.log.info("Post-Transform: S3 to DataSource...")

        # Note: We do not need to onboard because PandasToData already onboarded


if __name__ == "__main__":
    """Exercise the CSVToDataSource Class"""
    import sys
    from pathlib import Path

    # Get the path to the dataset in the repository data directory
    data_path = str(Path(sys.modules["workbench"].__file__).parent.parent.parent / "data" / "abalone.csv")

    # Create my Data Loader
    my_loader = CSVToDataSource(data_path, "abalone_data")
    my_loader.set_output_tags(["abalone", "public"])

    # Store this data as a Workbench DataSource
    my_loader.transform()



"""Experimental: Outlier detection based on nearest neighbors."""

from typing import Union
import pandas as pd


# Stub for the NearestNeighbors class
def get_neighbors(self, query_id: Union[int, str], include_self: bool = False):
    """
    Return neighbors of the given query ID, either by fixed neighbors or above a similarity threshold.

    Args:
        query_id (Union[int, str]): The ID of the query point.
        include_self (bool): Whether to include the query ID itself in the neighbor results.
    """
    return []


def compute_outliers(df: pd.DataFrame, target: str, id_column: str, model_type="classifier"):
    """
    Identifies outliers based on the target values of nearest neighbors.

    Args:
        model_type (str): Type of model to use for outlier detection. Defaults to "classifier".

    - For classification, an outlier is a point whose target differs from most of its neighbors.
    - For regression, an outlier is a point more than 3 standard deviations away from the mean of its neighbors.
    """
    if target is None:
        raise ValueError("Target column must be set to compute outliers.")

    outliers = []
    for idx in df.index:
        neighbors = get_neighbors(df.at[idx, id_column], include_self=False)

        if model_type == "classifier":
            majority_class = neighbors[target].mode()[0]
            is_outlier = df.at[idx, target] != majority_class

        else:  # Regression
            neighbor_mean = neighbors[target].mean()
            neighbor_std = neighbors[target].std()
            is_outlier = abs(df.at[idx, target] - neighbor_mean) > 3 * neighbor_std

        outliers.append(is_outlier)

    # Add the outlier column to the DataFrame
    df["outlier"] = outliers


if __name__ == "__main__":

    # Test the compute_outliers method
    compute_outliers(model_type="regressor")
    print(df)

    # Test the compute_outliers method with classification
    data = {
        "ID": ["a", "b", "c", "d", "e"],
        "Feature1": [0.1, 0.2, 0.3, 0.4, 0.5],
        "Feature2": [0.5, 0.4, 0.3, 0.2, 0.1],
        "target": ["red", "red", "red", "red", "blue"],
    }
    df = pd.DataFrame(data)
    prox = Proximity(df, id_column="ID", features=["Feature1", "Feature2"], target="target", n_neighbors=10)
    print(prox.all_neighbors())

    prox.compute_outliers(model_type="classifier")
    print(prox.neighbors(query_id="a"))



from dash import html, Dash, Output, Input
import dash_bootstrap_components as dbc
import logging
import socket


# Workbench Imports
from workbench.web_interface.components.plugin_interface import PluginInterface, PluginInputType
from workbench.api import DataSource, FeatureSet, Model, Endpoint, Meta
from workbench.api.compound import Compound
from workbench.api.graph_store import GraphStore
from workbench.api.pipeline import Pipeline
from workbench.utils.theme_manager import ThemeManager

# Setup Logging
log = logging.getLogger("workbench")


class PluginUnitTest:
    def __init__(self, plugin_class, theme="midnight_blue", input_data=None, auto_update=True, **kwargs):
        """A class to unit test a PluginInterface class.

        Args:
            plugin_class (PluginInterface): The PluginInterface class to test
            theme (str): The theme to use for the Dash app (default: "dark")
            input_data (Optional): The input data for this plugin (FeatureSet, Model, Endpoint, or DataFrame)
            auto_update (bool): Whether to automatically update the plugin properties (default: True)
            **kwargs: Additional keyword arguments
        """
        assert issubclass(
            plugin_class, PluginInterface
        ), "Plugin class has not passed all the PluginInterface validations"

        # If the input data is provided, let's store it for when update_properties is called
        self.input_data = input_data
        self.kwargs = kwargs

        # Set up the Theme Manager
        tm = ThemeManager()
        tm.set_theme(theme)
        css_files = tm.css_files()

        # Create the Dash app with the theme CSS files
        self.app = Dash(
            __name__,
            title="Plugin Unit Test",
            external_stylesheets=css_files,
        )
        self.port = 8050

        # Register the CSS route in the ThemeManager
        tm.register_css_route(self.app)

        # Instantiate the plugin
        self.plugin = plugin_class()
        self.component = self.plugin.create_component(f"{self.plugin.__class__.__name__.lower()}_test")

        # Set up the layout
        layout_children = [
            dbc.Row(self.component, style={"marginBottom": "20px", "height": "700px"}),
            html.Button("Update Plugin", id="update-button", style={"marginBottom": "20px"}),
        ]

        # Signal output displays
        layout_children.append(html.H3("Signals:"))
        for component_id, property in self.plugin.signals:
            # A Row with the component ID and property and an output div
            layout_children.append(html.H4(f"Property: {property}"))
            layout_children.append(html.Div(id=f"test-output-{component_id}-{property}"))

        # Set the layout
        self.app.layout = html.Div(
            [
                dbc.Container(layout_children, fluid=True, className="dbc dbc-ag-grid"),
            ],
            style={"margin": "40px"},
            **{"data-bs-theme": tm.data_bs_theme()},
        )

        # Make sure the plugin has a properties attribute (non-empty list of tuples)
        assert hasattr(self.plugin, "properties"), "Plugin must have a 'properties' attribute"
        if not self.plugin.properties:
            raise ValueError("Plugin must have a non-empty 'properties' attribute")

        # Call the internal method to update properties
        if auto_update:
            self._trigger_update()

        # Set up the test callback for updating the plugin
        @self.app.callback(
            [Output(component_id, property) for component_id, property in self.plugin.properties],
            [Input("update-button", "n_clicks")],
        )
        def update_plugin_properties(n_clicks):
            print("Update Plugin Properties")
            return self._trigger_update()

        # Set up callbacks for displaying output signals
        for component_id, property in self.plugin.signals:

            @self.app.callback(
                Output(f"test-output-{component_id}-{property}", "children"), Input(component_id, property)
            )
            def display_output_signal(signal_value):
                return f"{signal_value}"

        # Now register any internal callbacks
        self.plugin.register_internal_callbacks()

    def _trigger_update(self):
        """Trigger an update for the plugin properties based on its input type."""
        plugin_input_type = self.plugin.plugin_input_type

        if plugin_input_type == PluginInputType.DATA_SOURCE:
            data_source = self.input_data if self.input_data is not None else DataSource("abalone_data")
            return self.plugin.update_properties(data_source, **self.kwargs)
        elif plugin_input_type == PluginInputType.FEATURE_SET:
            feature_set = self.input_data if self.input_data is not None else FeatureSet("abalone_features")
            return self.plugin.update_properties(feature_set, **self.kwargs)
        elif plugin_input_type == PluginInputType.MODEL:
            model = self.input_data if self.input_data is not None else Model("abalone-regression")
            return self.plugin.update_properties(model, inference_run="auto_inference", **self.kwargs)
        elif plugin_input_type == PluginInputType.ENDPOINT:
            endpoint = self.input_data if self.input_data is not None else Endpoint("abalone-regression")
            return self.plugin.update_properties(endpoint, **self.kwargs)
        elif plugin_input_type == PluginInputType.PIPELINE:
            pipeline = self.input_data if self.input_data is not None else Pipeline("abalone_pipeline_v1")
            return self.plugin.update_properties(pipeline, **self.kwargs)
        elif plugin_input_type == PluginInputType.GRAPH:
            graph = self.input_data if self.input_data is not None else GraphStore().get("test/karate_club")
            return self.plugin.update_properties(graph, **self.kwargs)
        elif plugin_input_type == PluginInputType.DATAFRAME:
            df = self.input_data if self.input_data is not None else Meta().models(details=True)
            return self.plugin.update_properties(df, **self.kwargs)
        elif plugin_input_type == PluginInputType.COMPOUND:
            fake_compound = Compound("AQSOL-0001")
            fake_compound.smiles = "CC(C)C1=CC=C(C=C1)C(=O)O"
            compound = self.input_data if self.input_data is not None else fake_compound
            return self.plugin.update_properties(compound, **self.kwargs)
        else:
            raise ValueError(f"Invalid test type: {plugin_input_type}")

        # Set up callbacks for displaying output signals
        for component_id, property in self.plugin.signals:

            @self.app.callback(
                Output(f"test-output-{component_id}-{property}", "children"), Input(component_id, property)
            )
            def display_output_signal(signal_value):
                return f"{signal_value}"

        # Now register any internal callbacks
        self.plugin.register_internal_callbacks()

    def run(self):
        """Run the Dash server for the plugin, handling common errors gracefully."""
        while self.is_port_in_use(self.port):
            log.info(f"Port {self.port} is in use. Trying the next one...")
            self.port += 1  # Increment the port number until an available one is found

        log.info(f"Starting Dash server on port {self.port}...")
        self.app.run(debug=True, use_reloader=False, port=self.port)

    @staticmethod
    def is_port_in_use(port):
        """Check if a port is in use."""
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
            return sock.connect_ex(("127.0.0.1", port)) == 0



"""Aggregation: Perform Row Aggregation on a DataFrame"""

import pandas as pd
import logging

# Workbench Logger
log = logging.getLogger("workbench")


def aggregate(df: pd.DataFrame, group_column: str, features: list = None) -> pd.DataFrame:
    """Aggregate Row of a DataFrame
    Args:
        df: Pandas DataFrame
        group_column: The column to aggregate/group on
        features: List of column names (numeric) to perform aggregation on (default: None)
    Returns:
        Pandas DataFrame with aggregated rows and averaged numeric columns
    """

    # If no features are given, indentify all numeric columns
    if features is None:
        features = [x for x in df.select_dtypes(include="number").columns.tolist() if not x.endswith("id")]
        log.info("No features given, auto identifying numeric columns...")
        log.info(f"{features}")

    # Sanity checks
    if not all(column in df.columns for column in features):
        log.critical("Some features are missing in the DataFrame")
        return df
    if df.empty:
        log.critical("DataFrame is empty")
        return df

    # Now aggregate the DataFrame
    log.info(f"Aggregating dataframe, averaging {features}...")

    # Subset the DataFrame to only the features and group column
    df = df[features + [group_column]].copy()
    df["group_count"] = 1

    # Define the aggregation methods for each column
    agg_methods = {feature: "mean" for feature in features}
    agg_methods["group_count"] = "size"

    # Group by the group column and perform the aggregation
    df = df.groupby(group_column).agg(agg_methods).reset_index()

    # Return the DataFrame
    return df


def test():
    """Test for the Row Aggregation Class"""
    # Set some pandas options
    pd.set_option("display.max_columns", None)
    pd.set_option("display.width", 1000)

    # Make some fake data
    data = {
        "ID": [
            "id_0",
            "id_0",
            "id_2",
            "id_3",
            "id_4",
            "id_5",
            "id_6",
            "id_7",
            "id_8",
            "id_9",
        ],
        "feat1": [1.0, 1.0, 1.1, 3.0, 4.0, 1.0, 1.0, 1.1, 3.0, 4.0],
        "feat2": [1.0, 1.0, 1.1, 3.0, 4.0, 1.0, 1.0, 1.1, 3.0, 4.0],
        "feat3": [0.1, 0.1, 0.2, 1.6, 2.5, 0.1, 0.1, 0.2, 1.6, 2.5],
        "price": [31, 60, 62, 40, 20, 31, 61, 60, 40, 20],
        "outlier_group": [
            "sample",
            "a_low",
            "sample",
            "b_high",
            "sample",
            "c_high",
            "sample",
            "d_low",
            "sample",
            "e_high",
        ],
    }
    data_df = pd.DataFrame(data)

    # Aggregate the DataFrame
    new_df = aggregate(data_df, group_column="outlier_group")

    # Output the DataFrame
    print(new_df)


if __name__ == "__main__":
    test()



from pprint import pprint
from workbench.api import Meta

# Create our Meta Class and get a list of our Endpoints
meta = Meta()
endpoint_df = meta.endpoints()
print(f"Number of Endpoints: {len(endpoint_df)}")
print(endpoint_df)

# Get more details data on the Endpoints
endpoint_names = endpoint_df["Name"].tolist()
for name in endpoint_names:
    pprint(meta.endpoint(name))



"""GraphStore: Storage for NetworkX Graphs"""

import networkx as nx
from datetime import datetime
from typing import Union, Optional
import pandas as pd

# Workbench Imports
from workbench.core.cloud_platform.aws.aws_graph_store import AWSGraphStore


class GraphStore(AWSGraphStore):
    """GraphStore: Storage for NetworkX Graphs

    Common Usage:
        ```python
        graph_store = GraphStore()

        # List all Graphs
        graph_store.list()

        # Add Graph
        g = nx.erdos_renyi_graph(5, 0.5)
        graph_store.upsert("/test/my_graph", my_graph)

        # Retrieve Graph
        g = graph_store.get("/test/my_graph")
        print(g)

        # Delete Graph
        graph_store.delete("/test/my_graph")
        ```
    """

    def __init__(self, path_prefix: Optional[str] = None):
        """GraphStore Init Method

        Args:
            path_prefix (Optional[str]): Path prefix for storage locations (Defaults to None)
        """

        # Initialize the SuperClass
        super().__init__(path_prefix=path_prefix)

    def summary(self) -> pd.DataFrame:
        """Provide a summary of all graphs in the store.

        Returns:
            pd.DataFrame: Summary DataFrame with location, size, and modified date.
        """
        return super().summary()

    def details(self) -> pd.DataFrame:
        """Return detailed metadata for all stored graphs.

        Returns:
            pd.DataFrame: DataFrame with details like location, size, and last modified date.
        """
        return super().details()

    def check(self, location: str) -> bool:
        """Check if a graph exists.

        Args:
            location (str): Logical location of the graph.

        Returns:
            bool: True if the graph exists, False otherwise.
        """
        return super().check(location)

    def get(self, location: str) -> Union[nx.Graph, None]:
        """Retrieve a NetworkX graph from AWS S3.

        Args:
            location (str): Logical location of the graph.

        Returns:
            Union[nx.Graph, None]: The retrieved graph or None if not found.
        """
        return super().get(location)

    def upsert(self, location: str, graph: nx.Graph):
        """Insert or update a NetworkX graph in AWS S3.

        Args:
            location (str): Logical location to store the graph.
            graph (nx.Graph): The NetworkX graph to store.
        """
        super().upsert(location, graph)

    def list(self) -> list:
        """List all graphs in the store.

        Returns:
            list: A list of all graph locations in the store.
        """
        return super().list()

    def last_modified(self, location: str) -> Union[datetime, None]:
        """Return the last modified date of a graph.

        Args:
            location (str): Logical location of the graph.

        Returns:
            Union[datetime, None]: Last modified datetime or None if not found.
        """
        return super().last_modified(location)

    def delete(self, location: str):
        """Delete a NetworkX graph from AWS S3.

        Args:
            location (str): Logical location of the graph to delete.
        """
        super().delete(location)


if __name__ == "__main__":
    """Exercise the GraphStore Class"""
    # Create an GraphStore instance
    graph_store = GraphStore()

    # Create a test graph
    G = nx.erdos_renyi_graph(5, 0.5)
    for u, v, d in G.edges(data=True):
        d["weight"] = 1.0

    # Store the graph
    graph_store.upsert("test/test_graph", G)

    # Get a summary of the graphs
    print("Graph Store Summary:")
    print(graph_store.summary())

    # Display detailed metadata
    print("Graph Store Details:")
    print(graph_store.details())

    # Last modified date
    print("Last modified:", graph_store.last_modified("test/test_graph"))

    # List all graphs
    print("List all graphs in the store:")
    print(graph_store.list())

    # Retrieve and display the graph
    retrieved_graph = graph_store.get("test/test_graph")
    print("Retrieved Graph Edges:", retrieved_graph.edges(data=True))

    # Check if the graph exists
    print("Graph exists:", graph_store.check("test/test_graph"))

    # Delete the graph
    graph_store.delete("test/test_graph")
    print("Graph exists after deletion:", graph_store.check("test/test_graph"))



import json
import time

# Get the boto3 session from the Workbench Account Clamp
from workbench.core.cloud_platform.aws.aws_account_clamp import AWSAccountClamp

session = AWSAccountClamp().boto3_session

# Create EventBridge client
eventbridge_client = session.client("events")

# Define the event bus name
event_bus_name = "workbench"


# Function to monitor events
def monitor_events():
    while True:
        try:
            response = eventbridge_client.list_rules(EventBusName=event_bus_name)
            print(f"Monitoring events from EventBus: {event_bus_name}")

            for rule in response.get("Rules", []):
                print(f"\nRule: {rule['Name']}")
                events_response = eventbridge_client.describe_rule(Name=rule["Name"], EventBusName=event_bus_name)
                print(json.dumps(events_response, indent=2))

            # Wait for a short period before checking again
            time.sleep(10)

        except Exception as e:
            print(f"An error occurred: {str(e)}")
            time.sleep(10)


if __name__ == "__main__":
    monitor_events()



"""ModelsPageView pulls Model metadata from the AWS Service Broker with Details Panels on each Model"""

import pandas as pd

# Workbench Imports
from workbench.web_interface.page_views.page_view import PageView
from workbench.cached.cached_meta import CachedMeta
from workbench.cached.cached_model import CachedModel
from workbench.utils.symbols import tag_symbols


class ModelsPageView(PageView):
    def __init__(self):
        """ModelsPageView pulls Model metadata and populates a Details Panel"""
        # Call SuperClass Initialization
        super().__init__()

        # CachedMeta object for Cloud Platform Metadata
        self.meta = CachedMeta()

        # Initialize the Models DataFrame
        self.models_df = None
        self.refresh()

    def refresh(self):
        """Refresh the model data from the Cloud Platform"""
        self.log.important("Calling model page view refresh()..")
        self.models_df = self.meta.models(details=True)

        # Drop some columns
        self.models_df.drop(columns=["Ver", "Status", "_aws_url"], inplace=True, errors="ignore")

        # Add Health Symbols to the Model Group Name
        if "Health" in self.models_df.columns:
            self.models_df["Health"] = self.models_df["Health"].map(lambda x: tag_symbols(x))

    def models(self) -> pd.DataFrame:
        """Get all the data that's useful for this view

        Returns:
            pd.DataFrame: DataFrame of the Models View Data
        """
        return self.models_df

    @staticmethod
    def model_details(self, model_name: str) -> (dict, None):
        """Get all the details for the given Model Name
        Args:
            model_name(str): The Name of the Model
        Returns:
            dict: The details for the given Model (or None if not found)
        """
        model = CachedModel(model_name)
        if not model.exists():
            return {"Status": "Not Found"}
        elif not model.ready():
            return {"health_tags": model.get_health_tags()}

        # Return the Model Details
        return model.details()


if __name__ == "__main__":
    # Exercising the ModelsPageView
    import time
    from pprint import pprint

    # Create the class and get the AWS Model details
    model_view = ModelsPageView()

    # List the Models
    print("ModelsSummary:")
    summary = model_view.models()
    print(summary.head())

    # Get the details for the first Model
    my_model_name = summary["Model Group"].iloc[0]
    print("\nModelDetails:")
    details = model_view.model_details(my_model_name)
    pprint(details)

    # Give any broker threads time to finish
    time.sleep(1)



from workbench.utils.chem_utils import canonicalize
import pandas as pd


def test_canonicalize():
    """Test that canonicalize resolves known SMILES to their expected canonical forms."""
    # Define test cases for canonicalization
    test_data = [
        {"id": "Acetylacetone", "smiles": "CC(=O)CC(=O)C", "expected": "CC(=O)CC(C)=O"},
        {"id": "Imidazole", "smiles": "c1cnc[nH]1", "expected": "c1c[nH]cn1"},
        {"id": "Pyridone", "smiles": "C1=CC=NC(=O)C=C1", "expected": "O=c1cccccn1"},
        {"id": "Guanidine", "smiles": "C(=N)N=C(N)N", "expected": "N=CN=C(N)N"},
        {"id": "Catechol", "smiles": "c1cc(c(cc1)O)O", "expected": "Oc1ccccc1O"},
        {"id": "Formamide", "smiles": "C(=O)N", "expected": "NC=O"},
        {"id": "Urea", "smiles": "C(=O)(N)N", "expected": "NC(N)=O"},
        {"id": "Phenol", "smiles": "c1ccc(cc1)O", "expected": "Oc1ccccc1"},
        {"id": "Lactic Acid", "smiles": "CC(O)C(=O)O", "expected": "CC(O)C(=O)O"},  # Same as input
        {"id": "Pyruvic Acid", "smiles": "CC(=O)C(=O)O", "expected": "CC(=O)C(=O)O"},  # Same as input
        {"id": "Nitromethane", "smiles": "C[N+](=O)[O-]", "expected": "C[N+](=O)[O-]"},  # Same as input
    ]

    # Convert test data to a DataFrame
    df = pd.DataFrame(test_data)

    # Perform canonicalization
    result_df = canonicalize(df)

    # Test that the results match the expected canonical SMILES
    for index, row in result_df.iterrows():
        assert row["smiles_canonical"] == row["expected"], f"Canonicalization failed for {row['id']}"

    print("All canonicalization tests passed!")


if __name__ == "__main__":
    # Run the test
    test_canonicalize()



import pandas as pd
from rdkit import Chem
from rdkit.Chem import PandasTools

# Workbench Imports
from workbench.api.df_store import DFStore
from workbench.utils.chem_utils import add_compound_tags, compute_morgan_fingerprints, project_fingerprints


def prep_sdf_file(filepath: str) -> pd.DataFrame:

    # Load SDF file directly into a DataFrame
    df = PandasTools.LoadSDF(filepath, smilesName="smiles", molColName="molecule", includeFingerprints=False)
    print(df.head())
    print(f"Loaded {len(df)} compounds from {filepath}")
    print(f"Columns: {df.columns}")
    df.rename(columns={"ID": "id"}, inplace=True)

    # Column Information
    """
    | `Formula`         | Chemical formula of the compound.                                                             |
    | `FW`              | Molecular weight (Formula Weight) of the compound.                                            |
    | `DSSTox_CID`      | Unique identifier from the Distributed Structure-Searchable Toxicity (DSSTox) database.       |                                               |
    | `ID`              | Internal identifier for the compound.                                                         |
    | `smiles`          | Simplified Molecular Input Line Entry System representation of the compound's structure.      |
    | `molecule`        | Molecular structure information (format may vary). 
    | `NR-AR`           | Nuclear receptor assay for androgen receptor.     
    | `SR-HSE`          | Stress response assay for heat shock element.  |
    | `SR-ARE`          | Stress response assay for antioxidant response element.                                       |
    | `NR-Aromatase`    | Nuclear receptor assay for aromatase enzyme.                                                  |
    | `NR-ER-LBD`       | Nuclear receptor assay for estrogen receptor ligand-binding domain.                           |
    | `NR-AhR`          | Nuclear receptor assay for aryl hydrocarbon receptor.                                         |
    | `SR-MMP`          | Stress response assay for mitochondrial membrane potential.                                   |
    | `NR-ER`           | Nuclear receptor assay for estrogen receptor.                                                 |
    | `NR-PPAR-gamma`   | Nuclear receptor assay for peroxisome proliferator-activated receptor gamma.                  |
    | `SR-p53`          | Stress response assay for p53 tumor suppressor protein.                                       |
    | `SR-ATAD5`        | Stress response assay for ATAD5 (ATPase family AAA domain-containing protein 5).              |
    | `NR-AR-LBD`       | Nuclear receptor assay for androgen receptor ligand-binding domain.                           |
    """

    # Convert Formula weight to numeric
    df["FW"] = pd.to_numeric(df["FW"], errors="coerce")

    # Full list of assay columns for toxicity detection (0/1/NaN) for values
    assay_cols = [
        "NR-AR",  # Nuclear Receptor - Androgen Receptor
        "SR-HSE",  # Stress Response - Heat Shock Element
        "SR-ARE",  # Stress Response - Antioxidant Response Element
        "NR-Aromatase",  # Nuclear Receptor - Aromatase
        "NR-ER-LBD",  # Nuclear Receptor - Estrogen Receptor (Ligand Binding Domain)
        "NR-AhR",  # Nuclear Receptor - Aryl Hydrocarbon Receptor
        "SR-MMP",  # Stress Response - Mitochondrial Membrane Potential
        "NR-ER",  # Nuclear Receptor - Estrogen Receptor
        "NR-PPAR-gamma",  # Nuclear Receptor - Peroxisome Proliferator-Activated Receptor Gamma
        "SR-p53",  # Stress Response - p53 Tumor Suppressor Protein
        "SR-ATAD5",  # Stress Response - ATPase Family AAA Domain-Containing Protein 5
        "NR-AR-LBD",  # Nuclear Receptor - Androgen Receptor (Ligand Binding Domain)
    ]

    # Convert to numeric, coercing errors to NaN
    df[assay_cols] = df[assay_cols].apply(pd.to_numeric, errors="coerce")

    # Set toxic_any to 1 if any of the toxicity columns has a 1
    df["toxic_any"] = (df[assay_cols] == 1).any(axis=1).astype(int)
    print(df["toxic_any"].value_counts(dropna=False))

    # Convert SMILES to RDKit molecule objects (vectorized)
    if "molecule" not in df.columns:
        df["molecule"] = df["smiles"].apply(Chem.MolFromSmiles)

    # Add Compound Tags
    df = add_compound_tags(df, mol_column="molecule")

    # Now we have a 'tag' column with a list of tags in it, let's add "tox21" to the tags if the compound is toxic
    df["tags"] = df.apply(lambda row: row["tags"] + ["tox21"] if row["toxic_any"] == 1 else row["tags"], axis=1)

    # Compute Fingerprints
    df = compute_morgan_fingerprints(df, radius=2)

    # Project Fingerprints to 2D space
    df = project_fingerprints(df, projection="UMAP")

    # df = project_fingerprints(df, projection="TSNE")
    # df.rename(columns={"x": "x_tsne", "y": "y_tsne"}, inplace=True)
    # Convert compound tags to string and check for substrings, then convert to 0/1
    # df["toxic_tag"] = df["tags"].astype(str).str.contains("toxic").astype(int)
    # df["druglike_tag"] = df["tags"].astype(str).str.contains("druglike").astype(int)

    # Drop the molecule column
    df.drop(columns=["molecule"], inplace=True)

    # Return the prepared DataFrame
    return df


if __name__ == "__main__":

    # Load the training and test sets
    training = "/Users/briford/data/workbench/tox21/training/tox21_10k_data_all.sdf"
    test = "/Users/briford/data/workbench/tox21/testing/tox21_10k_challenge_test.sdf"
    final_test = "/Users/briford/data/workbench/tox21/testing/tox21_10k_challenge_score.sdf"

    # Process the data and put it into the DFStore
    df_store = DFStore()
    df = prep_sdf_file(training)
    df_store.upsert("/datasets/chem_info/tox21", df)

    # Okay, so the test datasets have a different format than the training dataset (so skip for now)
    """
    df = prep_sdf_file(test)
    df_store.upsert("/datasets/chem_info/tox21_test", df)
    df = prep_sdf_file(final_test)
    df_store.upsert("/datasets/chem_info/tox21_final_test", df)
    """



"""AWSAccountClamp provides logic/functionality over a set of AWS IAM Services"""

import boto3
from botocore.exceptions import (
    ClientError,
    UnauthorizedSSOTokenError,
    TokenRetrievalError,
)
from botocore.client import BaseClient
import logging
from sagemaker.session import Session as SageSession

# Workbench Imports
from workbench.core.cloud_platform.aws.aws_session import AWSSession
from workbench.utils.config_manager import ConfigManager, FatalConfigError


class AWSAccountClamp:
    """AWSAccountClamp: Singleton class for connecting to an AWS Account"""

    _instance = None  # Class attribute to hold the singleton instance

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super(AWSAccountClamp, cls).__new__(cls)
        return cls._instance

    def __init__(self):
        """AWSAccountClamp Initialization"""
        if hasattr(self, "_initialized") and self._initialized:
            return  # Prevent reinitialization

        # Setup code (only runs once)
        self.log = logging.getLogger("workbench")
        self.log.info("Creating the AWSAccountClamp Singleton...")

        # ConfigManager and AWS setup
        self.cm = ConfigManager()
        if not self.cm.config_okay():
            self.log.error("Workbench Configuration Incomplete...")
            self.log.error("Run the 'workbench' command and follow the prompts...")
            raise FatalConfigError()
        self.workbench_bucket_name = self.cm.get_config("WORKBENCH_BUCKET")
        self.aws_session = AWSSession()
        self.boto3_session = self.aws_session.boto3_session

        # Check our caller/base AWS Identity
        try:
            self.account_id = boto3.client("sts").get_caller_identity()["Account"]
            self.region = boto3.session.Session().region_name
        except (ClientError, UnauthorizedSSOTokenError, TokenRetrievalError):
            self.log.critical("AWS Identity Check Failure: Check AWS_PROFILE and/or Renew SSO Token...")
            raise FatalConfigError()

        # Check our Assume Role
        self.log.info("Checking Workbench Assumed Role...")
        self.aws_session.assumed_role_info()

        # Check our Workbench API Key and Load the License
        self.log.info("Checking Workbench API License...")
        self.cm.load_and_check_license(self.account_id)
        self.cm.print_license_info()

        # Mark the instance as initialized
        self._initialized = True

    def check_aws_identity(self) -> bool:
        """Check the Caller/Base AWS Identity currently active (not the assumed role)"""

        # Using the caller/base boto3 client (not the assumed role session)
        try:
            sts = boto3.client("sts")
            sts.get_caller_identity()
            return True
        except (ClientError, UnauthorizedSSOTokenError):
            msg = "AWS Identity Check Failure: Check AWS_PROFILE and/or Renew SSO Token..."
            self.log.critical(msg)
            raise RuntimeError(msg)

    def check_assumed_role(self) -> bool:
        """Check the AWS Identity of the Assumed Role"""
        try:
            self.aws_session.assumed_role_info()
            return True
        except (ClientError, UnauthorizedSSOTokenError):
            msg = "AWS Identity Check Failure: Check AWS_PROFILE and/or Renew SSO Token..."
            self.log.critical(msg)
            raise RuntimeError(msg)

    def get_aws_account_info(self) -> dict:
        """Get the AWS Account Information

        Returns:
            dict: A dictionary of the AWS Account Information
        """
        info = {}
        sts = boto3.client("sts")
        try:
            identity = sts.get_caller_identity()
            info["Account"] = identity["Account"]
            info["IdentityArn"] = identity["Arn"]
            info["Region"] = self.region
            return info
        except (ClientError, UnauthorizedSSOTokenError):
            self.log.critical("AWS Identity Check Failure: Check AWS_PROFILE and/or Renew SSO Token...")
            return info

    def check_workbench_bucket(self) -> bool:
        """Check if the Workbench S3 Bucket exists and is accessible"""
        s3 = self.boto3_session.client("s3")  # Use client, not resource
        try:
            s3.head_bucket(Bucket=self.workbench_bucket_name)
            self.log.info(f"The {self.workbench_bucket_name} bucket exists and is accessible")
            return True
        except ClientError as e:
            error_code = e.response["Error"]["Code"]
            if error_code == "404":
                self.log.critical(f"The {self.workbench_bucket_name} bucket does not exist")
            elif error_code == "403":
                self.log.critical(f"Access denied to {self.workbench_bucket_name} bucket")
            else:
                self.log.error(f"Error checking S3 bucket: {e}")
            return False

    def sagemaker_session(self) -> "SageSession":
        """Create a workbench SageMaker session (using our boto3 refreshable session)

        Returns:
            SageSession: A SageMaker session object
        """
        return SageSession(boto_session=self.boto3_session)

    def sagemaker_client(self) -> BaseClient:
        """Create a workbench SageMaker client (using our boto3 refreshable session)

        Returns:
            BaseClient: A SageMaker client object
        """
        return self.boto3_session.client("sagemaker")


if __name__ == "__main__":
    """Exercise the AWS Account Clamp Class"""

    # Create the class
    aws_account_clamp = AWSAccountClamp()

    # Check out that AWS Account Clamp is working AOK
    """Check if the AWS Account is Setup Correctly"""
    print("\n\n*** AWS Caller/Base Identity Check ***")
    aws_account_clamp.check_aws_identity()
    print("Caller/Base Identity Check Success...")

    print("\n\n*** AWS Assumed Role Check ***")
    aws_account_clamp.check_assumed_role()
    print("Assumed Role Check Success...")

    print("\n\n*** AWS Workbench Bucket Check ***")
    aws_account_clamp.check_workbench_bucket()
    print("Workbench Bucket Check Success...")

    print("\n\n*** AWS Sagemaker Session/Client Check ***")
    sm_client = aws_account_clamp.sagemaker_client()
    print(sm_client.list_feature_groups()["FeatureGroupSummaries"])



import logging
from concurrent.futures import ThreadPoolExecutor, as_completed
from workbench.api import FeatureSet
from workbench.utils.aws_utils import list_tags_with_throttle

# Set up logging
log = logging.getLogger()
log.setLevel(logging.INFO)
log.addHandler(logging.StreamHandler())


def spam_list_tags(arn, sm_session, index):
    log.info(f"Spamming list_tags_with_throttle: {index}")
    return list_tags_with_throttle(arn, sm_session)


if __name__ == "__main__":

    # Grab our test FeatureSet
    my_features = FeatureSet("abalone_features")
    arn = my_features.arn()
    sm_session = my_features.sm_session

    # Test on an ARN that doesn't exist
    log.info("Testing on an ARN that doesn't exist")
    tags = list_tags_with_throttle("arn:aws:sagemaker:us-west-2:123456789012:fake-arn", sm_session)

    # Run spamming in parallel to induce throttling
    log.info("Spamming list_tags_with_throttle")
    with ThreadPoolExecutor(max_workers=10) as executor:
        futures = [executor.submit(spam_list_tags, arn, sm_session, i) for i in range(1000)]

        # Collect results
        for future in as_completed(futures):
            try:
                tags = future.result()
            except Exception as e:
                log.error(f"Encountered an error: {e}")



import botocore
import logging
import sys
import time
import functools
import json
import base64
import re
import os
from typing import Union, List, Callable, Optional
import pandas as pd
import awswrangler as wr
from awswrangler.exceptions import NoFilesFound
from pathlib import Path
import posixpath
from botocore.exceptions import ClientError
from sagemaker.session import Session as SageSession
from collections.abc import Mapping, Iterable


# Workbench Imports
from workbench.utils.config_manager import ConfigManager
from workbench.utils.deprecated_utils import deprecated

# Workbench Logger
log = logging.getLogger("workbench")
cm = ConfigManager()


def client_error_printout(err: botocore.exceptions.ClientError):
    """Helper method to get information about a botocore.exceptions.ClientError"""
    error_code = err.response["Error"]["Code"]
    error_message = err.response["Error"]["Message"]
    operation_name = err.operation_name
    service_name = err.response["ResponseMetadata"].get("ServiceName", "Unknown")
    request_id = err.response["ResponseMetadata"]["RequestId"]

    # Output the error information
    log.error(f"Error Code: {error_code}")
    log.error(f"Error Message: {error_message}")
    log.error(f"Operation Name: {operation_name}")
    log.error(f"Service Name: {service_name}")
    log.error(f"Request ID: {request_id}")


def aws_throttle(func=None, retry_intervals=None):
    """
    Decorator to handle AWS throttling exceptions with exponential backoff.

    Args:
        func: This is a decorator detail just ignore it.
        retry_intervals (list[int], optional): List of intervals in seconds between retries.
                                               If None, defaults to exponential backoff.
    """
    if func is None:
        return lambda f: aws_throttle(f, retry_intervals=retry_intervals)

    # This is currently commented out (we might want to use it later)
    # service_hold_time = 2  # Seconds to wait before calling AWS function
    default_intervals = [2**i for i in range(1, 9)]  # Default exponential backoff: 2, 4, 8... 256 seconds
    intervals = retry_intervals or default_intervals

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        for attempt, delay in enumerate(intervals, start=1):
            try:
                # Add sleep before calling AWS func if running as a service
                # if cm.running_as_service:
                #    time.sleep(service_hold_time)
                return func(*args, **kwargs)
            except ClientError as e:
                if e.response["Error"]["Code"] == "ThrottlingException":
                    log_level = log.critical if delay > 100 else log.error if delay > 30 else log.warning
                    log_level(f"{func.__name__}: ThrottlingException ({attempt}): Retrying in {delay} seconds...")
                    time.sleep(delay)
                else:
                    raise
        # If we exhaust all retries, raise an exception
        raise ClientError(
            {
                "Error": {
                    "Code": "ThrottlingException",
                    "Message": f"{func.__name__} failed after {len(intervals)} retries",
                }
            },
            func.__name__,
        )

    return wrapper


def not_found_returns_none(func: Optional[Callable] = None, *, resource_name: str = "AWS resource") -> Callable:
    """Decorator to handle AWS resource not found (returns None) and re-raising otherwise.

    Args:
        func (Callable, optional): The function being decorated.
        resource_name (str): Name of the AWS resource being accessed. Used for clearer error messages.
    """
    not_found_errors = {
        "ResourceNotFound",
        "ResourceNotFoundException",
        "EntityNotFoundException",
        "ValidationException",
        "NoSuchBucket",
    }

    def decorator(inner_func: Callable) -> Callable:
        @functools.wraps(inner_func)
        def wrapper(*args, **kwargs):
            try:
                return inner_func(*args, **kwargs)
            except ClientError as error:
                error_code = error.response["Error"]["Code"]
                if error_code in not_found_errors:
                    log.warning(f"{resource_name} not found: {error_code}, returning None...")
                    return None
                else:
                    log.critical(f"Critical error in AWS call: {error_code}")
                    raise
            except wr.exceptions.NoFilesFound:
                log.info(f"Resource {resource_name} not found returning None...")
                return None

        return wrapper

    # If func is None, the decorator was called with arguments
    if func is None:
        return decorator
    else:
        # If func is not None, the decorator was used without arguments
        return decorator(func)


@deprecated(version="0.9")
def list_tags_with_throttle(arn: str, sm_session) -> dict:
    """A Wrapper around SageMaker's list_tags method that handles throttling
    Args:
        arn (str): The ARN of the SageMaker resource
        sm_session (SageSession): A SageMaker session object
    Returns:
        dict: A dictionary of tags

    Note: AWS List Tags can get grumpy if called too often
    """

    # Log the call
    log.debug(f"Calling list_tags for {arn}...")
    sleep_times = [2, 4, 8, 16, 32, 64]
    max_attempts = len(sleep_times)

    # Sanity check the ARN
    if arn is None:
        log.error("Called list_tags_with_throttle(arn==None)!")
        return {}

    # Loop with exponential backoff
    for attempt in range(max_attempts):
        try:
            # Call the AWS List Tags method
            aws_tags = sm_session.list_tags(arn)
            meta = aws_tags_to_dict(aws_tags)  # Convert the AWS Tags to a dictionary
            return meta

        except botocore.exceptions.ClientError as e:
            error_code = e.response["Error"]["Code"]
            error_message = e.response["Error"]["Message"]

            # Check for ThrottlingException
            if error_code == "ThrottlingException":
                if attempt < 2:
                    log.info(f"ThrottlingException ({attempt}): list_tags on {arn}")
                elif attempt < 4:
                    log.warning(f"ThrottlingException ({attempt}): list_tags on {arn}")
                else:
                    log.error(f"ThrottlingException ({attempt}): list_tags on {arn}")

            # Check specific (Not Found) exceptions
            elif error_code in ["ValidationException", "ResourceNotFoundException"]:
                log.warning(f"Probably Fine -- {arn} AWS Validation/NotFound Exception: {error_code} - {error_message}")
                return {}
            else:
                # Handle other ClientErrors that may occur
                log.error(f"ClientError: {error_code} - {error_message}")
                raise e

            # Sleep for a bit before trying again
            time.sleep(sleep_times[attempt])

    # If we get here, we've failed to retrieve the tags
    log.error(f"Failed to retrieve tags for {arn}!")
    return {}


@deprecated(version="0.9")
def workbench_meta_from_catalog_table_meta(table_meta: dict) -> dict:
    """Retrieve the Workbench metadata from AWS Data Catalog table metadata
    Args:
        table_meta (dict): The AWS Data Catalog table metadata
    Returns:
        dict: The Workbench metadata that's stored in the Parameters field
    """
    # Get the Parameters field from the table metadata
    params = table_meta.get("Parameters", {})
    return {key: decode_value(value) for key, value in params.items() if "workbench" in key}


def decode_value(value):
    # Try to base64 decode the value
    try:
        value = base64.b64decode(value).decode("utf-8")
    except Exception:
        pass
    # Try to JSON decode the value
    try:
        value = json.loads(value)
    except Exception:
        pass

    # Okay, just return whatever we have
    return value


def is_valid_tag(tag):
    pattern = r"^([a-zA-Z0-9_.:/=+\-@]*)$"
    return re.match(pattern, tag) is not None


def dict_to_aws_tags(meta_data: dict) -> list:
    """AWS Tags are in an odd format, so we need to convert, encode, and chunk the data
    Args:
        meta_data (dict): Dictionary of metadata to convert to AWS Tags
    """
    chunked_data = {}  # Store any chunked data here
    chunked_keys = []  # Store any keys to remove here

    # AWS Tag Storage has the following constraints:
    # - 256 char maximum per tag
    # - 50 tags maximum per artifact (model/endpoint/etc)
    char_limit = 256
    max_tags = 50

    # Loop through the data: Convert non-string values to JSON strings, and chunk large data
    output_data = {}
    current_tags = len(meta_data.keys())
    for key, value in meta_data.items():
        # Convert data to JSON string
        if not isinstance(value, str):
            value = json.dumps(value, separators=(",", ":"))

        # Make sure the value is valid
        if not is_valid_tag(value):
            log.important(f"Base64 encoding metadata: {value}")
            value = base64.b64encode(value.encode()).decode()

        # Check if the value will fit in the 256-character limit
        if len(value) < char_limit:
            output_data[key] = value

        # If the value is longer than 256 but we have room to split it into chunks
        elif len(value) < (char_limit * (max_tags - current_tags)):
            log.important(f"Chunking metadata for key {key} with length {len(value)}...")
            chunked_keys.append(key)
            chunks = _chunk_data(key, value)
            for chunk in chunks:
                chunked_data[chunk] = chunks[chunk]
                current_tags += 1

        # Too long to store in AWS Tags
        else:
            log.error(f"Metadata for key '{key}' is large {len(value)} and cannot be stored in AWS Tags!")

    # Now add the chunked data to the output data
    output_data.update(chunked_data)
    log.info(f"Processed tags has {len(output_data.keys())} keys...")

    # Now convert to AWS Tags format
    aws_tags = []
    for key, value in output_data.items():
        aws_tags.append({"Key": key, "Value": value})
    return aws_tags


def aws_tags_to_dict(aws_tags) -> dict:
    """Internal: AWS Tags are in an odd format, so convert, decode, and de-chunk"""

    # Stitch together any chunked data
    stitched_data = {}
    regular_tags = {}

    for item in aws_tags:
        key = item["Key"]
        value = item["Value"]

        # Check if this key is a chunk
        if "_chunk_" in key:
            base_key, chunk_num = key.rsplit("_chunk_", 1)

            if base_key not in stitched_data:
                stitched_data[base_key] = {}

            stitched_data[base_key][int(chunk_num)] = value
        else:
            regular_tags[key] = decode_value(value)

    # Stitch chunks back together and decode
    for base_key, chunks in stitched_data.items():
        # Sort by chunk number and concatenate
        sorted_chunks = [chunks[i] for i in sorted(chunks.keys())]
        stitched_base64_str = "".join(sorted_chunks)

        # Decode the stitched base64 string
        try:
            stitched_json_str = base64.b64decode(stitched_base64_str).decode("utf-8")
        except UnicodeDecodeError:
            stitched_json_str = stitched_base64_str
        try:
            stitched_dict = json.loads(stitched_json_str)
        except json.decoder.JSONDecodeError:
            stitched_dict = stitched_json_str

        regular_tags[base_key] = stitched_dict

    return regular_tags


def parse_sagemaker_arn(arn: str) -> tuple:
    """Parse a SageMaker ARN to extract the resource type and name.

    Args:
        arn (str): The ARN of the SageMaker resource.

    Returns:
        tuple: A tuple containing the resource type and resource name.
    """
    parts = arn.split(":")
    resource_part = parts[-1]  # "resource-type/resource-name"
    resource_type, resource_name = resource_part.split("/", 1)
    return resource_type, resource_name


def _chunk_data(base_key: str, data: str) -> dict:
    # Initialize variables
    chunk_size = 256  # Max size for AWS tag value
    chunks = {}

    # Split the data into chunks with numbered 'chunk_' keys
    for i in range(0, len(data), chunk_size):
        chunk = data[i : i + chunk_size]
        chunks[f"{base_key}_chunk_{i // chunk_size + 1}"] = chunk

    return chunks


def extract_data_source_basename(source: str) -> str:
    """Extract the basename from a data source
    Args:
        source (str): The data source
    Returns:
        str: The basename of the data source
    """

    # If the source is a Pandas DataFrame, return 'dataframe'
    if isinstance(source, pd.DataFrame):
        return "dataframe"

    # Convert PosixPath to string if necessary
    if isinstance(source, Path):
        source = str(source)

    # Check if the source is an S3 path
    if source.startswith("s3://"):
        basename = posixpath.basename(source)
        name_without_extension = os.path.splitext(basename)[0]
        return name_without_extension + "_data"

    # Check if the source is a local file path
    elif os.path.isfile(source):
        basename = os.path.basename(source)
        name_without_extension = os.path.splitext(basename)[0]
        return name_without_extension + "_data"

    # If it's neither, assume it's already a data source name
    else:
        return source


def list_s3_files(s3_path: str, extensions: str = "*.csv") -> List[str]:
    """
    Lists files in an S3 path with specified extension.

    Args:
    s3_path (str): The full S3 path (e.g., 's3://my-bucket/my-prefix/').
    extensions (str): File extension to filter by, defaults to '*.csv'.

    Returns:
    List[str]: A list of file paths matching the extension in the S3 path.
    """
    files = wr.s3.list_objects(path=s3_path, suffix=extensions.lstrip("*"))
    return files


def newest_path(s3_locations: list[str], sm_session: SageSession) -> Union[str, None]:
    """Determine which S3 bucket and prefix combination has the newest files.

    Args:
        s3_locations (list[str]): A list of full S3 bucket and prefix combinations.
        sm_session (SageSession): A SageMaker session object.

    Returns:
        str: The full S3 bucket and prefix combination with the newest files, or None if no files are found.
    """
    s3_client = sm_session.boto_session.client("s3")
    newest_location, latest_time = None, None

    for location in s3_locations:
        bucket, prefix = location.replace("s3://", "", 1).split("/", 1)
        response = s3_client.list_objects_v2(Bucket=bucket, Prefix=prefix)

        for file in response.get("Contents", []):
            if latest_time is None or file["LastModified"] > latest_time:
                newest_location, latest_time = location, file["LastModified"]

    return newest_location


def pull_s3_data(s3_path: str, embedded_index=False) -> Union[pd.DataFrame, None]:
    """Helper method to pull data from S3 storage

    Args:
        s3_path (str): S3 Path to the Artifact
        embedded_index (bool, optional): Is the index embedded in the CSV? Defaults to False.

    Returns:
        pd.DataFrame: DataFrame of the Artifact (metrics, CM, regression_preds) (might be None)
    """

    # Sanity check for undefined S3 paths (None)
    if s3_path.startswith("None"):
        return None

    # Pull the CSV file from S3
    try:
        if embedded_index:
            df = wr.s3.read_csv(s3_path, index_col=0)
        else:
            df = wr.s3.read_csv(s3_path)
        return df
    except NoFilesFound:
        log.info(f"Could not find S3 data at {s3_path}...")
        return None
    except Exception as e:
        log.error(f"Failed to pull data from {s3_path}!")
        log.error(e)
        return None


@deprecated(version="0.9", stack_trace=True)
def compute_size(obj: object) -> int:
    """Recursively calculate the size of an object including its contents.

    Args:
        obj (object): The object whose size is to be computed.

    Returns:
        int: The total size of the object in bytes.
    """
    if isinstance(obj, Mapping):
        return sys.getsizeof(obj) + sum(compute_size(k) + compute_size(v) for k, v in obj.items())
    elif isinstance(obj, Iterable) and not isinstance(obj, (str, bytes, bytearray)):
        return sys.getsizeof(obj) + sum(compute_size(item) for item in obj)
    else:
        return sys.getsizeof(obj)


@deprecated(version="0.9")
def num_columns_ds(data_info):
    """Helper: Compute the number of columns from the storage descriptor data"""
    try:
        return len(data_info["StorageDescriptor"]["Columns"])
    except KeyError:
        return "-"


@deprecated(version="0.9")
def num_columns_fs(data_info):
    """Helper: Compute the number of columns from the feature group data"""
    try:
        return len(data_info["FeatureDefinitions"])
    except KeyError:
        return "-"


@deprecated(version="0.9")
def aws_url(artifact_info, artifact_type, aws_account_clamp):
    """Helper: Try to extract the AWS URL from the Artifact Info Object"""
    if artifact_type == "S3":
        # Construct the AWS URL for the S3 Bucket
        name = artifact_info["Name"]
        region = aws_account_clamp.region
        s3_prefix = f"incoming-data/{name}"
        bucket_name = aws_account_clamp.workbench_bucket_name
        base_url = "https://s3.console.aws.amazon.com/s3/object"
        return f"{base_url}/{bucket_name}?region={region}&prefix={s3_prefix}"
    elif artifact_type == "GlueJob":
        # Construct the AWS URL for the Glue Job
        region = aws_account_clamp.region
        job_name = artifact_info["Name"]
        base_url = f"https://{region}.console.aws.amazon.com/gluestudio/home"
        return f"{base_url}?region={region}#/editor/job/{job_name}/details"
    elif artifact_type == "DataSource":
        details = artifact_info.get("Parameters", {}).get("workbench_details", "{}")
        return json.loads(details).get("aws_url", "unknown")
    elif artifact_type == "FeatureSet":
        aws_url = artifact_info.get("workbench_meta", {}).get("aws_url", "unknown")
        # Hack for constraints on the SageMaker Feature Group Tags
        return aws_url.replace("__question__", "?").replace("__pound__", "#")


if __name__ == "__main__":
    """Exercise the AWS Utils"""
    from pprint import pprint
    from workbench.core.artifacts.feature_set_core import FeatureSetCore
    from workbench.core.cloud_platform.aws.aws_account_clamp import AWSAccountClamp

    # Grab out SageMaker Session from the AWS Account Clamp
    sm_session = AWSAccountClamp().sagemaker_session()
    boto3_session = AWSAccountClamp().boto3_session

    # Get the Sagemaker client from the AWS Account Clamp
    sm_client = AWSAccountClamp().sagemaker_client()

    @aws_throttle
    def list_sagemaker_models():
        response = sm_client.list_models()
        return response["Models"]

    print(list_sagemaker_models())

    # Test the aws_throttle decorator with ThrottlingExceptions
    # @aws_throttle
    # def test_throttling():
    #     raise ClientError({"Error": {"Code": "ThrottlingException"}}, "test")
    # test_throttling()

    @not_found_returns_none
    def test_not_found():
        raise ClientError({"Error": {"Code": "ResourceNotFoundException"}}, "test")

    test_not_found()

    @not_found_returns_none(resource_name="my_not_found_resource")
    def test_not_found():
        raise ClientError({"Error": {"Code": "ResourceNotFoundException"}}, "test")

    test_not_found()

    try:

        @not_found_returns_none
        def test_other_error():
            # Raise a different error to test the error handler
            raise ClientError({"Error": {"Code": "SomeOtherError"}}, "test")

        test_other_error()
    except ClientError:
        print("AOK Expected Error :)")

    # Test a FeatureSetCore object (that uses some of these methods)
    my_features = FeatureSetCore("test_features")
    my_meta = my_features.workbench_meta()
    pprint(my_meta)

    # Test listing_files in an S3 folder method
    print(list_s3_files("s3://workbench-public-data/common"))

    # Test the newest files in an S3 folder method
    s3_path = "s3://sandbox-sageworks-artifacts/endpoints/inference/abalone-regression"
    most_recent = newest_path([s3_path], sm_session)

    # Add a health tag
    my_features.add_health_tag("needs_onboard")
    print(my_features.get_health_tags())

    # Add a user tag
    my_features.add_tag("test_tag")
    my_tags = my_features.get_tags()
    pprint(my_tags)

    # Add workbench meta data
    my_features.upsert_workbench_meta({"test_meta": "test_value"})
    my_meta = my_features.workbench_meta()
    pprint(my_meta)

    # Test adding a None value
    my_features.upsert_workbench_meta({"test_meta": None})
    my_meta = my_features.workbench_meta()
    pprint(my_meta)

    # Add workbench meta data (testing regular expression constraints)
    my_features.upsert_workbench_meta({"test_meta": "test_{:value"})
    my_meta = my_features.workbench_meta()
    pprint(my_meta)

    # Add non-string workbench meta data
    my_features.upsert_workbench_meta({"test_meta": {"foo": "bar"}})
    my_meta = my_features.workbench_meta()
    pprint(my_meta)

    # Add some large workbench meta data (string)
    large_data = "x" * 512
    my_features.upsert_workbench_meta({"large_meta": large_data})
    my_meta = my_features.workbench_meta()
    pprint(my_meta)

    # Add some large workbench meta data (dict)
    large_data = {"data": "x" * 512, "more_data": "y" * 512}
    my_features.upsert_workbench_meta({"large_meta": large_data})
    my_meta = my_features.workbench_meta()
    pprint(my_meta)

    # Remove the tag
    my_features.delete_metadata("large_meta")
    my_meta = my_features.workbench_meta()
    pprint(my_meta)

    # Test the list_tags_with_throttle method
    arn = my_features.arn()
    tags = list_tags_with_throttle(arn, sm_session)
    pprint(tags)



"""Cache class for key/value pairs"""

import time
from collections import OrderedDict
import atexit


class Cache:
    """In process memory cache. Not thread safe.
    Usage:
         cache = Cache(max_size=5, expire=10)
         cache.set('foo', 'bar')
         cache.get('foo')
         > bar
         time.sleep(11)
         cache.get('foo')
         > None
         cache.clear()
    """

    def __init__(self, max_size=10000, expire=None, prefix="", postfix=""):
        """Cache Initialization"""
        self.store = OrderedDict()
        self.max_size = max_size
        self.expire = expire
        self._compression_timer = 600
        self._last_compression = time.time()
        self.prefix = prefix if not prefix or prefix.endswith(":") else prefix + ":"
        self.postfix = postfix if not postfix or postfix.startswith(":") else ":" + postfix

        # Try to do cleanup/serialization at exit
        atexit.register(self.cleanup)

    def _get_full_key(self, key):
        return f"{self.prefix}{key}{self.postfix}"

    @staticmethod
    def check():
        """Check the status of this cache"""
        return True  # I'm doing great, thanks for asking

    def set(self, key, value):
        """Add an item to the cache
        Args:
               key: item key
               value: the value associated with this key
        """
        self._check_limit()
        actual_key = self._get_full_key(key)
        _expire = time.time() + self.expire if self.expire else None
        self.store[actual_key] = (value, _expire)

    def get(self, key):
        """Get an item from the cache
        Args:
            key: item key
        Returns:
            the value of the item or None if the item isn't in the cache
        """
        actual_key = self._get_full_key(key)
        data = self.store.get(actual_key)
        if not data:
            return None
        value, expire = data
        if expire and time.time() > expire:
            del self.store[actual_key]
            return None
        return value

    def delete(self, key):
        # Delete the key and its full key (if either exists)
        if key in self.store:
            del self.store[key]
        full_key = self._get_full_key(key)
        if full_key in self.store:
            del self.store[full_key]

    def list_keys(self):
        """List all keys in the cache"""
        return list(self.store.keys())

    def list_subkeys(self, key):
        """List all sub-keys in the cache"""
        return [k for k in self.store.keys() if k.startswith(f"{self.prefix}{key}")]

    def clear(self):
        """Clear the cache"""
        self.store = OrderedDict()

    def dump(self):
        """Dump the cache (for debugging)"""
        for key in self.store.keys():
            print(key, ":", self.store.get(key))

    @property
    def size(self):
        return len(self.store)

    def cleanup(self):
        """Cleanup the cache (if we need to)"""
        pass

    def _check_limit(self):
        """Internal method: check if current cache size exceeds maximum cache
        size and pop the oldest item in this case"""

        # First compress
        self._compress()

        # Then check the max size
        if len(self.store) >= self.max_size:
            self.store.popitem(last=False)  # FIFO

    def _compress(self):
        """Internal method to compress the cache. This method will
        expire any old items in the cache, making the cache smaller"""

        # Don't compress too often
        now = time.time()
        if self._last_compression + self._compression_timer < now:
            self._last_compression = now
            for key in list(self.store.keys()):
                self.get(key)

    def __repr__(self):
        return f"MemoryCache(Prefix={self.prefix})"


if __name__ == "__main__":
    """Exercise the Cache class"""

    # Create the Cache
    my_cache = Cache(max_size=5, expire=1)
    my_cache.set("foo", "bar")

    # Test storage
    assert my_cache.get("foo") == "bar"

    # Test expire
    time.sleep(1.1)
    assert my_cache.get("foo") is None

    # Test max_size
    my_cache = Cache(max_size=5)
    for i in range(6):
        my_cache.set(str(i), i)

    # So the '0' key should no longer be there FIFO
    assert my_cache.get("0") is None
    assert my_cache.get("5") is not None

    # Test listing keys
    print("Listing Keys...")
    print(my_cache.list_keys())
    print(my_cache.list_subkeys("foo"))

    # Make sure size is working
    assert my_cache.size == 5

    # Dump the cache
    my_cache.dump()

    # Test deleting
    my_cache.delete("1")
    assert my_cache.get("1") is None

    # Test storing 'null' values
    my_cache.set(0, "foo")
    my_cache.set(0, "bar")
    my_cache.set(None, "foo")
    my_cache.set("", None)
    assert my_cache.get("") is None
    assert my_cache.get(None) == "foo"
    assert my_cache.get(0) == "bar"

    # Test the cache compression
    my_cache = Cache(max_size=5, expire=1)
    for i in range(5):
        my_cache.set(str(i), i)
    my_cache._compression_timer = 1
    assert my_cache.size == 5

    # Make sure compression is working
    time.sleep(1.1)
    my_cache._compress()
    assert my_cache.size == 0

    # Also make sure compression call is throttled
    my_cache._compress()  # Should not output a compression message

    # Test prefix/postfix
    my_cache = Cache(max_size=5, prefix="pre", postfix="post")
    my_cache.set("foo", "bar")
    assert my_cache.get("foo") == "bar"
    my_cache.dump()



"""A Correlation Matrix component"""

from dash import dcc
import plotly.graph_objects as go
import pandas as pd

# Workbench Imports
from workbench.web_interface.components.component_interface import ComponentInterface
from workbench.utils.theme_manager import ThemeManager
from workbench.utils.color_utils import remove_middle_colors


# This class is basically a specialized version of a Plotly Heatmap
# For heatmaps see (https://plotly.com/python/heatmaps/)
class CorrelationMatrix(ComponentInterface):
    """Correlation Matrix Component"""

    def __init__(self):
        """Initialize the Regression Plot Class"""

        # Initialize the Theme Manager
        self.theme_manager = ThemeManager()

        # Hack the colorscale (revisit this)
        colorscale = self.theme_manager.colorscale("diverging")
        self.colorscale = remove_middle_colors(colorscale)

        # Call the parent class constructor
        super().__init__()

    def create_component(self, component_id: str) -> dcc.Graph:
        """Create a Correlation Matrix Component without any data.
        Args:
            component_id (str): The ID of the web component
        Returns:
            dcc.Graph: The Correlation Matrix Component
        """
        return dcc.Graph(id=component_id, figure=self.display_text("Waiting for Data..."))

    def update_properties(self, data_source_details: dict) -> go.Figure:
        """Create a Correlation Matrix Figure for the numeric columns in the dataframe.

        Args:
            data_source_details (dict): A dictionary containing DataSource details.

        Returns:
            plotly.graph_objs.Figure: A Figure object containing the correlation matrix.
        """

        # Sanity check the data
        if data_source_details is None:
            return self.display_text("No Details Data Found", figure_height=200)
        if "column_stats" not in data_source_details:
            return self.display_text("No column_stats Found", figure_height=200)

        # Threshold for the correlation matrix (filter out low values)
        threshold = 0.4

        # Convert the data details into a correlation dataframe
        df = self._corr_df_from_data_details(data_source_details, threshold=threshold)

        # If the dataframe is empty then return a message
        if df.empty:
            return self.display_text("No Correlations Found", figure_height=200)

        # Okay so the heatmap has inverse y-axis ordering, so we need to flip the dataframe
        df = df.iloc[::-1]

        # Okay so there are numerous issues with getting back the index of the clicked on point
        # so we're going to store the indexes of the columns (this is SO stupid)
        x_labels = [f"{c}:{i}" for i, c in enumerate(df.columns)]
        y_labels = [f"{c}:{i}" for i, c in enumerate(df.index)]

        # Create the heatmap plot with custom settings
        height = max(350, len(df.index) * 50)
        fig = go.Figure(
            data=go.Heatmap(
                z=df,
                x=x_labels,
                y=y_labels,
                xgap=2,  # Add space between cells
                ygap=2,
                name="",
                colorscale=self.colorscale,
                zmin=-1,
                zmax=1,
            )
        )
        fig.update_layout(margin={"t": 10, "b": 10, "r": 10, "l": 10, "pad": 0}, height=height)

        # Truncate labels to a maximum of 24 characters
        x_axes_labels = [f"{c[:20]}..." if len(c) > 24 else c for c in df.columns]
        y_axes_labels = [f"{c[:20]}..." if len(c) > 24 else c for c in df.index]
        fig.update_xaxes(tickvals=x_labels, ticktext=x_axes_labels, tickangle=30, showgrid=False)
        fig.update_yaxes(tickvals=y_labels, ticktext=y_axes_labels, showgrid=False)

        # Now we're going to customize the annotations and filter out low values
        for i, row in enumerate(df.index):
            for j, col in enumerate(df.columns):
                value = df.loc[row, col]
                if abs(value) > threshold:
                    fig.add_annotation(x=j, y=i, text=f"{value:.2f}", showarrow=False)

        return fig

    @staticmethod
    def _corr_df_from_data_details(data_details: dict, threshold: float = 0.4) -> pd.DataFrame:
        """Internal: Create a Pandas DataFrame in the form given by df.corr() from DataSource details.
        Args:
            data_details (dict): A dictionary containing DataSource details.
            threshold (float): Any correlations below this value will be excluded.
        Returns:
            pd.DataFrame: A Pandas DataFrame containing the correlation matrix
        """

        # Sanity check
        if not data_details:
            return pd.DataFrame()

        # Process the data so that we can make a Dataframe of the correlation data
        column_stats = data_details["column_stats"]
        corr_dict = {key: info["correlations"] for key, info in column_stats.items() if "correlations" in info}
        corr_df = pd.DataFrame(corr_dict)

        # The diagonal will be NaN, so fill it with 0
        corr_df.fillna(0, inplace=True)

        # Now filter out any correlations below the threshold
        corr_df = corr_df.loc[:, (corr_df.abs().max() > threshold)]
        corr_df = corr_df[(corr_df.abs().max(axis=1) > threshold)]

        # If the correlation matrix is bigger than 12x12 then we need to filter it down
        while corr_df.shape[0] > 12:
            # Now filter out any correlations below the threshold
            corr_df = corr_df.loc[:, (corr_df.abs().max() > threshold)]
            corr_df = corr_df[(corr_df.abs().max(axis=1) > threshold)]
            threshold += 0.05

        # Return the correlation dataframe in the form of df.corr()
        corr_df.sort_index(inplace=True)
        corr_df = corr_df[corr_df.index]
        return corr_df


if __name__ == "__main__":
    # This class takes in data details and generates a Correlation Matrix
    from workbench.api.data_source import DataSource

    tm = ThemeManager()
    tm.set_theme("dark")

    ds = DataSource("test_data")
    ds_details = ds.details()

    # Instantiate the CorrelationMatrix class
    corr_plot = CorrelationMatrix()

    # Generate the figure
    fig = corr_plot.update_properties(ds_details)

    # Apply dark theme
    fig.update_layout()

    # Show the figure
    fig.show()



"""An abstract class that defines the web component interface for Workbench"""

import logging
from abc import ABC, abstractmethod
from typing import Any, Union
import traceback
import re
from functools import wraps
import plotly.graph_objects as go
import pandas as pd
from dash import dcc
from dash.development.base_component import Component
import networkx as nx

# Workbench Imports
from workbench.api import DataSource, FeatureSet, Model, Endpoint
from workbench.api.compound import Compound
from workbench.cached.cached_data_source import CachedDataSource
from workbench.cached.cached_feature_set import CachedFeatureSet
from workbench.cached.cached_model import CachedModel
from workbench.cached.cached_endpoint import CachedEndpoint

from workbench.api.pipeline import Pipeline

log = logging.getLogger("workbench")


def get_reversed_stack_trace(exception, limit=5):
    # Get the full stack trace
    full_stack = traceback.format_exception(type(exception), exception, exception.__traceback__)
    # Reverse the stack and take the last `limit` lines
    relevant_stack = full_stack[-limit:]
    return "".join(relevant_stack)


class ComponentInterface(ABC):
    """A Abstract Web Component Interface
    Notes:
      - The 'create_container' method create a gcc.Graph, html.Div, etc
      - The 'update_properties' method generates the property values
    """

    log = logging.getLogger("workbench")

    # Define the Workbench Objects Type (revisit allowing the non-cached objects)
    WorkbenchObject = Union[
        DataSource,
        FeatureSet,
        Model,
        Endpoint,
        Pipeline,
        nx.Graph,
        pd.DataFrame,
        CachedDataSource,
        CachedFeatureSet,
        CachedModel,
        CachedEndpoint,
        Compound,
    ]

    def __init__(self, theme: str = "DARK"):
        """Initialize the Component Interface

        Args:
            theme (str): The theme to use for the component ("LIGHT" or "DARK" default is "DARK")
        """
        self.component_id = None
        self.container = None
        self.properties = []
        self.signals = []

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)

        # Automatically apply the error handling decorator to the create_component and update_properties methods
        if hasattr(cls, "create_component") and callable(cls.create_component):
            cls.create_component = create_component_handler(cls.create_component)
        if hasattr(cls, "update_properties") and callable(cls.update_properties):
            cls.update_properties = update_properties_handler(cls.update_properties)

    @abstractmethod
    def create_component(self, component_id: str, **kwargs: Any) -> Component:
        """Create a Dash Component/Container without any data.

        Args:
            component_id (str): The ID of the web component
            kwargs (Any): Any additional arguments to pass to the component

        Returns:
           Component: A Dash Base Component
        """
        pass

    def update_properties(self, data_object: WorkbenchObject) -> list:
        """Update the properties of the component/container

        Args:
            data_object (workbench_object/dataframe): A Workbench object or DataFrame

        Returns:
            list: A list of the updated property values for the component
        """
        pass

    def generate_component_id(self) -> str:
        """This helper method returns the component ID for the component
        Returns:
            str: An auto generated component ID
        """

        # Get the plugin class name
        plugin_class_name = self.__class__.__name__

        # Convert the plugin class name to snake case component ID
        component_id = re.sub("([a-z0-9])([A-Z])", r"\1_\2", plugin_class_name).lower()
        return component_id

    @staticmethod
    def display_text(text_message: str, figure_height: int = None, font_size=24) -> go.Figure:
        """This helper method displays a text message figure for the component
        Args:
            text_message (str): The text message to display
            figure_height (int): The height of the figure (default: None)
            font_size (int): The font size of the message (default: 32)
        Returns:
            go.Figure: A Plotly Figure
        """
        text_display_text = go.Figure()

        # If the text message has any \n characters, replace them with <br> for HTML
        text_message = text_message.replace("\n", "<br>")

        text_display_text.add_annotation(
            x=0.5,
            y=0.5,
            xref="paper",
            yref="paper",
            text=text_message,
            showarrow=False,
            font=dict(size=font_size, color="#9999cc"),
        )

        layout_options = dict(
            xaxis=dict(showticklabels=False, zeroline=False, showgrid=False),
            yaxis=dict(showticklabels=False, zeroline=False, showgrid=False),
            margin=dict(l=0, r=0, b=0, t=0),
        )

        if figure_height is not None:
            layout_options["height"] = figure_height

        text_display_text.update_layout(**layout_options)

        return text_display_text


# These are helper decorators to catch errors in plugin methods
def create_component_handler(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as e:
            # Get the class name of the plugin
            class_name = args[0].__class__.__name__ if args else "UnknownPlugin"

            # Generate our stack trace figure
            figure, _ = stack_trace_figure(class_name, e)
            return dcc.Graph(id="error", figure=figure)

    return wrapper


def update_properties_handler(func):
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        try:
            return func(self, *args, **kwargs)
        except Exception as e:
            # Get the class name of the plugin
            class_name = self.__class__.__name__

            # Generate our stack trace figure
            figure, error_info = stack_trace_figure(class_name, e)

            # Prepare the error output to match the properties format
            error_output = []
            for component_id, property in self.properties:
                if property == "figure":
                    error_output.append(figure)
                elif property in ["children", "value", "data"]:
                    error_output.append(error_info)
                elif property == "columnDefs":
                    error_output.append([{"headerName": "Crash", "field": "Crash"}])
                elif property == "rowData":
                    error_output.append([{"Crash": error_info}])
                else:
                    error_output.append(None)  # Fallback for other properties
            return error_output

    return wrapper


def stack_trace_figure(class_name: str, e: Exception) -> tuple[go.Figure, str]:
    """This helper method returns a Plotly Figure and the UI error info with the stack trace of an exception
    Args:
        class_name (str): The class name of the plugin
        e (Exception): The exception object
    Returns:
        tuple: A tuple containing the Plotly Figure and the stack trace error info
    """

    # Get the reversed stack trace
    reversed_stack_trace = get_reversed_stack_trace(e)
    header = f"{class_name} Crashed"
    stack_output = f"{header}: \n{reversed_stack_trace}"
    log.critical(stack_output)

    # Now split lines that are too long
    max_len = 80
    wrapped_stack_output = "\n".join(
        [line[i : i + max_len] for line in stack_output.split("\n") for i in range(0, len(line), max_len)]
    )

    # Send the wrapped stack output as a figure with text
    figure = ComponentInterface.display_text(wrapped_stack_output, figure_height=400, font_size=12)

    return figure, stack_output  # Return both figure and the stack trace



"""Convert training views from 0/1 to FALSE/TRUE in the specified AWS Glue database.

Note: This script is a 'schema' change for the training views, and is a 'one-time'
    operation. The code quality here is not as important as the correctness of the
    operation and since this will only be run once for existing clients and never
    again, we don't want to sweat the details.
"""

import json
import re
import base64
import logging
import awswrangler as wr
from workbench.core.cloud_platform.aws.aws_account_clamp import AWSAccountClamp

log = logging.getLogger("workbench")

# Initialize your AWS session and Glue client
aws_account_clamp = AWSAccountClamp()
session = aws_account_clamp.boto3_session
glue_client = session.client("glue")


def _decode_view_sql(encoded_sql: str) -> str:
    """Decode the base64-encoded SQL query from the view.

    Args:
        encoded_sql (str): The encoded SQL query in the ViewOriginalText.

    Returns:
        str: The decoded SQL query.
    """
    # Extract the base64-encoded content from the comment
    match = re.search(r"Presto View: ([\w=+/]+)", encoded_sql)
    if match:
        base64_sql = match.group(1)
        decoded_bytes = base64.b64decode(base64_sql)
        decoded_str = decoded_bytes.decode("utf-8")

        # Parse the decoded string as JSON to extract the SQL
        try:
            view_json = json.loads(decoded_str)
            return view_json.get("originalSql", "")
        except json.JSONDecodeError:
            log.error("Failed to parse the decoded view SQL as JSON.")
            return ""
    return ""


def convert_training_views(database):
    """Convert training views from 0/1 to FALSE/TRUE in the specified AWS Glue database"""

    # Use the Glue client to get the list of tables (views) from the database
    paginator = glue_client.get_paginator("get_tables")

    for page in paginator.paginate(DatabaseName=database):
        for table in page["TableList"]:
            # Check if the table name ends with "_training" and is a view
            if table["Name"].endswith("_training") and table.get("TableType") == "VIRTUAL_VIEW":
                print(f"Checking view: {table['Name']}...")

                # Decode the 'ViewOriginalText' for the view
                view_original_text = _decode_view_sql(table.get("ViewOriginalText"))
                if view_original_text and (" THEN 0" in view_original_text or " THEN 1" in view_original_text):
                    print(f"\tConverting view: {table['Name']}...")

                    # Update the THEN and ELSE view definitions by replacing 0/1 with FALSE/TRUE
                    updated_query = view_original_text.replace(" THEN 0", " THEN FALSE").replace(
                        " THEN 1", " THEN TRUE"
                    )
                    updated_query = updated_query.replace(" ELSE 0", " ELSE FALSE").replace(" ELSE 1", " ELSE TRUE")

                    # Construct the full CREATE OR REPLACE VIEW query
                    query = f"""
                    CREATE OR REPLACE VIEW {table['Name']} AS
                    {updated_query}
                    """

                    try:
                        # Execute the query using awswrangler
                        query_execution_id = wr.athena.start_query_execution(
                            sql=query,
                            database=database,
                            boto3_session=session,
                        )
                        print(f"\tQueryExecutionId: {query_execution_id}")

                        # Wait for the query to complete
                        wr.athena.wait_query(query_execution_id=query_execution_id, boto3_session=session)
                        print(f"\tSuccessfully converted view: {table['Name']}")
                    except Exception as e:
                        print(f"\tError updating view {table['Name']}: {e}")
                else:
                    print(f"\tNo conversion needed for view: {table['Name']}")


if __name__ == "__main__":
    # Specify your database scope
    database_scope = ["sagemaker_featurestore"]
    for db in database_scope:
        convert_training_views(db)



"""S3HeavyToDataSource: Class to move HEAVY S3 Files into a Workbench DataSource"""

import sys
import boto3
from botocore.exceptions import ClientError

# Import all the AWS Glue Python Libraries
from awsglue.utils import getResolvedOptions
from awsglue.context import GlueContext
from awsglue.job import Job
from pyspark.context import SparkContext
from awsglue.dynamicframe import DynamicFrame
from awsglue.transforms import Relationalize
from pyspark.sql.functions import col, to_timestamp


class S3HeavyToDataSource:
    def __init__(self, glue_context: GlueContext, input_name: str, output_name: str):
        """S3HeavyToDataSource: Class to move HEAVY S3 Files into a Workbench DataSource

        Args:
            glue_context: GlueContext, AWS Glue Specific wrapper around SparkContext
            input_name (str): The S3 Path to the files to be loaded
            output_name (str): The Name of the Workbench DataSource to be created
        """
        self.log = glue_context.get_logger()

        # FIXME: Pull these from Parameter Store or Config
        self.input_name = input_name
        self.output_name = output_name
        self.output_meta = {"workbench_input": self.input_name}
        workbench_bucket = "s3://sandbox-workbench-artifacts"
        self.data_sources_s3_path = workbench_bucket + "/data-sources"

        # Our Spark Context
        self.glue_context = glue_context

    @staticmethod
    def resolve_choice_fields(dyf):
        # Get schema fields
        schema_fields = dyf.schema().fields

        # Collect choice fields
        choice_fields = [(field.name, "cast:long") for field in schema_fields if field.dataType.typeName() == "choice"]
        print(f"Choice Fields: {choice_fields}")

        # If there are choice fields, resolve them
        if choice_fields:
            dyf = dyf.resolveChoice(specs=choice_fields)

        return dyf

    def timestamp_conversions(self, dyf: DynamicFrame, time_columns: list = []) -> DynamicFrame:
        """Convert columns in the DynamicFrame to the correct data types
        Args:
            dyf (DynamicFrame): The DynamicFrame to convert
            time_columns (list): A list of column names to convert to timestamp
        Returns:
            DynamicFrame: The converted DynamicFrame
        """

        # Convert the timestamp columns to timestamp types
        spark_df = dyf.toDF()
        for column in time_columns:
            spark_df = spark_df.withColumn(column, to_timestamp(col(column)))

        # Convert the Spark DataFrame back to a Glue DynamicFrame and return
        return DynamicFrame.fromDF(spark_df, self.glue_context, "output_dyf")

    @staticmethod
    def remove_periods_from_columns(dyf: DynamicFrame) -> DynamicFrame:
        """Remove periods from column names in the DynamicFrame
        Args:
            dyf (DynamicFrame): The DynamicFrame to convert
        Returns:
            DynamicFrame: The converted DynamicFrame
        """
        # Extract the column names from the schema
        old_column_names = [field.name for field in dyf.schema().fields]

        # Create a new list of renamed column names
        new_column_names = [name.replace(".", "_") for name in old_column_names]
        print(old_column_names)
        print(new_column_names)

        # Create a new DynamicFrame with renamed columns
        for c_old, c_new in zip(old_column_names, new_column_names):
            dyf = dyf.rename_field(f"`{c_old}`", c_new)
        return dyf

    def transform(
        self,
        input_type: str = "json",
        timestamp_columns: list = None,
        output_format: str = "parquet",
    ):
        """Convert the CSV or JSON data into Parquet Format in the Workbench S3 Bucket, and
        store the information about the data to the AWS Data Catalog workbench database
        Args:
            input_type (str): The type of input files, either 'csv' or 'json'
            timestamp_columns (list): A list of column names to convert to timestamp
            output_format (str): The format of the output files, either 'parquet' or 'orc'
        """

        # Add some tags here
        tags = ["heavy"]

        # Create the Output Parquet file S3 Storage Path
        s3_storage_path = f"{self.data_sources_s3_path}/{self.output_name}"

        # Read JSONL files from S3 and infer schema dynamically
        self.log.info(f"Reading JSONL files from {self.input_name}...")
        input_dyf = self.glue_context.create_dynamic_frame.from_options(
            connection_type="s3",
            connection_options={
                "paths": [self.input_name],
                "recurse": True,
                "gzip": True,
            },
            format=input_type,
            # format_options={'jsonPath': 'auto'}, Look into this later
        )
        self.log.info("Incoming DataFrame...")
        input_dyf.show(5)
        input_dyf.printSchema()

        # Resolve Choice fields
        resolved_dyf = self.resolve_choice_fields(input_dyf)

        # The next couple of lines of code is for un-nesting any nested JSON
        # Create a Dynamic Frame Collection (dfc)
        dfc = Relationalize.apply(resolved_dyf, name="root")

        # Aggregate the collection into a single dynamic frame
        output_dyf = dfc.select("root")

        print("Before TimeStamp Conversions")
        output_dyf.printSchema()

        # Convert any timestamp columns
        output_dyf = self.timestamp_conversions(output_dyf, timestamp_columns)

        # Relationalize will put periods in the column names. This will cause
        # problems later when we try to create a FeatureSet from this DataSource
        output_dyf = self.remove_periods_from_columns(output_dyf)

        print("After TimeStamp Conversions and Removing Periods from column names")
        output_dyf.printSchema()

        # Write Parquet files to S3
        self.log.info(f"Writing Parquet files to {s3_storage_path}...")
        self.glue_context.purge_s3_path(s3_storage_path, {"retentionPeriod": 0})
        self.glue_context.write_dynamic_frame.from_options(
            frame=output_dyf,
            connection_type="s3",
            connection_options={
                "path": s3_storage_path
                # "partitionKeys": ["year", "month", "day"],
            },
            format=output_format,
        )

        # Set up our Workbench metadata (description, tags, etc)
        description = f"Workbench data source: {self.output_name}"
        workbench_meta = {"workbench_tags": self.tag_delimiter.join(tags)}
        for key, value in self.output_meta.items():
            workbench_meta[key] = value

        # Create a new table in the AWS Data Catalog
        self.log.info(f"Creating Data Catalog Table: {self.output_name}...")

        # Converting the Spark Types to Athena Types
        def to_athena_type(col):
            athena_type_map = {"long": "bigint"}
            spark_type = col.dataType.typeName()
            return athena_type_map.get(spark_type, spark_type)

        column_name_types = [{"Name": col.name, "Type": to_athena_type(col)} for col in output_dyf.schema().fields]

        # Our parameters for the Glue Data Catalog are different for Parquet and ORC
        if output_format == "parquet":
            glue_input_format = "org.apache.hadoop.hive.ql.io.parquet.MapredParquetInputFormat"
            glue_output_format = "org.apache.hadoop.hive.ql.io.parquet.MapredParquetOutputFormat"
            serialization_library = "org.apache.hadoop.hive.ql.io.parquet.serde.ParquetHiveSerDe"
        else:
            glue_input_format = "org.apache.hadoop.hive.ql.io.orc.OrcInputFormat"
            glue_output_format = "org.apache.hadoop.hive.ql.io.orc.OrcInputFormat"
            serialization_library = "org.apache.hadoop.hive.ql.io.orc.OrcSerde"

        table_input = {
            "Name": self.output_name,
            "Description": description,
            "Parameters": workbench_meta,
            "TableType": "EXTERNAL_TABLE",
            "StorageDescriptor": {
                "Columns": column_name_types,
                "Location": s3_storage_path,
                "InputFormat": glue_input_format,
                "OutputFormat": glue_output_format,
                "Compressed": True,
                "SerdeInfo": {
                    "SerializationLibrary": serialization_library,
                },
            },
        }

        # Delete the Data Catalog Table if it already exists
        glue_client = boto3.client("glue")
        try:
            glue_client.delete_table(DatabaseName="workbench", Name=self.output_name)
            self.log.info(f"Deleting Data Catalog Table: {self.output_name}...")
        except ClientError as e:
            if e.response["Error"]["Code"] != "EntityNotFoundException":
                raise e

        self.log.info(f"Creating Data Catalog Table: {self.output_name}...")
        glue_client.create_table(DatabaseName="workbench", TableInput=table_input)

        # All done!
        self.log.info(f"{self.input_name} --> {self.output_name} complete!")


if __name__ == "__main__":
    """Glue Job for the S3HeavyToDataSource Class"""

    # Get the arguments for this Glue Job
    args = getResolvedOptions(sys.argv, ["JOB_NAME"])

    # Grab the SparkContext, GlueContext, and Job
    # These are all AWS Glue Job specific
    sc = SparkContext()
    glueContext = GlueContext(sc)
    job = Job(glueContext)
    job.init(args["JOB_NAME"], args)

    # Test the Heavy Data Loader
    input_path = "s3://sandbox-workbench-artifacts/incoming-data/dns/"
    data_output_name = "heavy_dns"
    my_loader = S3HeavyToDataSource(glueContext, input_path, data_output_name)

    # Store this data as a Workbench DataSource
    my_loader.transform(timestamp_columns=["timestamp"])

    # Commit the Glue Job
    job.commit()



# Quote from AWS Docs: https://docs.aws.amazon.com/athena/latest/ug/tables-databases-columns-names.html
"""
Use lower case for table names and table column names in Athena

Athena accepts mixed case in DDL and DML queries, but lower cases the names when it
executes the query. For this reason, avoid using mixed case for table or column names,
and do not rely on casing alone in Athena to distinguish such names. For example,
if you use a DDL statement to create a column named Castle, the column created
will be lowercased to castle.
"""


# Workbench Imports
from workbench.core.cloud_platform.aws.aws_account_clamp import AWSAccountClamp

boto3_session = AWSAccountClamp().boto3_session


# Create a Glue client and Athena client
glue_client = boto3_session.client("glue")
athena_client = boto3_session.client("athena")

# Define the database, table name, and S3 path
database_name = "workbench"
table_name = "test_table"
s3_path = "s3://test-bucket/test/"

# Define the original column names with mixed case
original_column_dict = {"Id": "int", "Name": "string", "Age": "int"}

# Define the Athena DDL query to create the table
ddl_query = f"""
CREATE EXTERNAL TABLE IF NOT EXISTS {database_name}.{table_name} (
    `Id` int,
    `Name` string,
    `Age` int
)
STORED AS PARQUET
LOCATION '{s3_path}'
"""

# Execute the DDL query
execution_id = athena_client.start_query_execution(
    QueryString=ddl_query,
    QueryExecutionContext={"Database": database_name},
    ResultConfiguration={"OutputLocation": "s3://your-output-bucket/path/"},
)["QueryExecutionId"]

# Wait for the query execution to complete (you may want to add a timeout or a more sophisticated waiting mechanism)
athena_client.get_query_execution(QueryExecutionId=execution_id)

# Use Glue client to read back the table (as Athena doesn't provide a direct way to get column names)
response = glue_client.get_table(DatabaseName=database_name, Name=table_name)

# Extract the column names from the response
retrieved_columns = [col["Name"] for col in response["Table"]["StorageDescriptor"]["Columns"]]

# Compare the original column names to the retrieved column names
original_columns = list(original_column_dict.keys())
print("\nOriginal Columns")
print(original_columns)
print("\nRetrieved Columns")
print(retrieved_columns)



"""SHAP Summary Plot visualization component for XGBoost models"""

from dash import dcc
import plotly.graph_objects as go
import pandas as pd
import numpy as np
from typing import Dict, List

# Workbench Imports
from workbench.cached.cached_model import CachedModel
from workbench.web_interface.components.plugin_interface import PluginInterface, PluginPage, PluginInputType
from workbench.utils.theme_manager import ThemeManager
from workbench.utils.plot_utils import beeswarm_offsets


class ShapSummaryPlot(PluginInterface):
    """SHAP Summary Visualization Component for XGBoost model explanations"""

    auto_load_page = PluginPage.MODEL
    plugin_input_type = PluginInputType.MODEL

    def __init__(self):
        """Initialize the ShapSummaryPlot plugin class"""
        self.component_id = None
        self.theme_manager = ThemeManager()
        super().__init__()

    def create_component(self, component_id: str) -> dcc.Graph:
        """Create a SHAP Summary Visualization Component without any data."""
        self.component_id = component_id
        self.container = dcc.Graph(
            id=component_id,
            figure=self.display_text("Waiting for SHAP data..."),
            config={"scrollZoom": False, "doubleClick": "reset", "displayModeBar": False},
        )
        self.properties = [(self.component_id, "figure")]
        self.signals = [(self.component_id, "clickData")]
        return self.container

    def update_properties(self, model: CachedModel, **kwargs) -> list:
        """Create a SHAP Summary Plot for feature importance visualization."""
        # Basic validation
        shap_data = model.shap_data()
        shap_sample_rows = model.shap_sample()
        if shap_data is None or shap_sample_rows is None:
            return [self.display_text("SHAP data not available")]

        # Check if the model is multiclass
        is_multiclass = isinstance(shap_data, dict)
        if is_multiclass:
            class_labels = model.class_labels()
            id_column = shap_data[list(shap_data.keys())[0]].columns[0]
            fig = self._create_multiclass_summary_plot(shap_data, shap_sample_rows, id_column, class_labels)

        # Regression or binary classification
        else:
            id_column = shap_data.columns[0]
            fig = self._create_summary_plot(shap_data, shap_sample_rows, id_column)
        return [fig]

    def _create_summary_plot(self, shap_df: pd.DataFrame, sample_df: pd.DataFrame, id_column: str) -> go.Figure:
        """Create a SHAP summary plot for a single class."""

        # Remove bias column if present
        if "bias" in shap_df.columns:
            shap_df.drop(columns=["bias"], inplace=True)

        # Grab the shap features (all columns except the ID column)
        shap_features = [feature for feature in shap_df.columns if feature != id_column]

        # Right now we are only supporting numeric features
        shap_features = [
            feature for feature in shap_features if feature in sample_df.select_dtypes(include="number").columns
        ]

        # Merge SHAP values with feature values
        merged_df = pd.merge(shap_df, sample_df, on=id_column, how="inner", suffixes=("_shap", ""))

        # Create figure
        fig = go.Figure()

        # Add a zero vertical line for reference
        fig.add_shape(
            type="line",
            x0=0,
            x1=0,
            y0=-0.5,
            y1=len(shap_features) - 0.5,
            line=dict(color="gray", width=2),
            layer="below",
        )

        # Add traces for each feature
        for i, feature in enumerate(shap_features):
            feature_shap = f"{feature}_shap"

            # Normalize feature values for this specific feature (0-1 scale)
            feature_vals = merged_df[feature].values
            norm_vals = (feature_vals - np.min(feature_vals)) / (np.max(feature_vals) - np.min(feature_vals) + 1e-10)

            # Get y positions with beeswarm offsets
            y_jitter = beeswarm_offsets(merged_df[feature_shap]) + i

            # Add scatter plot
            fig.add_trace(
                go.Scatter(
                    x=merged_df[feature_shap],
                    y=y_jitter,
                    mode="markers",
                    name=feature,
                    marker=dict(
                        color=norm_vals,
                        colorbar=(
                            dict(
                                title="Feature Value",  # Add the title
                                title_side="right",
                                tickvals=[0, 1],
                                ticktext=["Low", "High"],
                                thickness=10,
                            )
                            if i == 0
                            else None
                        ),  # Only show colorbar for first feature
                        opacity=1.0,
                        size=8,
                        showscale=(i == 0),  # Only show color scale for first feature
                    ),
                    showlegend=False,
                    hoverinfo="text",
                    hovertext=[
                        f"Feature: {feature}<br>SHAP value: {shap:.4f}<br>Feature value: {val:.4f}"
                        for shap, val in zip(merged_df[feature_shap], feature_vals)
                    ],
                )
            )

        # Update layout
        tick_labels = [f[:15] for f in shap_features]  # Truncate labels to 15 characters
        fig.update_layout(
            title="SHAP Summary Plot: Feature Impact",
            xaxis_title="SHAP Value (Impact on Model Output)",
            margin={"t": 50, "b": 60, "r": 0, "l": 0, "pad": 0},
            height=max(400, 50 * len(shap_features)),
            plot_bgcolor=self.theme_manager.background(),
            xaxis=dict(showgrid=False, zeroline=False),
            yaxis=dict(
                tickvals=list(range(len(shap_features))),
                ticktext=tick_labels,
                autorange="reversed",  # Most important features at top
                showgrid=False,
                zeroline=False,
            ),
        )
        return fig

    def _create_multiclass_summary_plot(
        self, shap_values: Dict[str, pd.DataFrame], sample_df: pd.DataFrame, id_column: str, class_labels: List[str]
    ) -> go.Figure:
        """Create a SHAP summary plot for multiple classes with class selector."""

        # Get list of classes
        class_names = class_labels
        first_class = class_names[0]

        # Create base figure for first class
        main_fig = self._create_summary_plot(shap_values[first_class], sample_df, id_column)

        # Store traces from first class
        first_class_traces = list(range(len(main_fig.data)))
        traces_per_class = len(first_class_traces)

        # Add traces for all other classes (initially invisible)
        for class_name in class_names[1:]:
            # Create figure for this class
            class_fig = self._create_summary_plot(shap_values[class_name], sample_df, id_column)

            # Add all traces from class_fig to main_fig (set to invisible)
            for trace in class_fig.data:
                # Update hover text to include class
                if hasattr(trace, "hovertext"):
                    new_hovertext = []
                    for ht in trace.hovertext:
                        new_hovertext.append(f"Class: {class_name}<br>{ht}")
                    trace.hovertext = new_hovertext

                # Set to invisible and add to main figure
                trace.visible = False
                main_fig.add_trace(trace)

        # Create buttons for class selection
        buttons = []
        for i, class_name in enumerate(class_names):
            # Create visibility list
            total_traces = len(main_fig.data)

            # Set visibility (True for selected class, False for others)
            visibility = [False] * total_traces
            class_traces_start = i * traces_per_class
            class_traces_end = class_traces_start + traces_per_class

            for j in range(class_traces_start, class_traces_end):
                visibility[j] = True

            # Add button for this class
            buttons.append(dict(method="restyle", label=str(class_name), args=[{"visible": visibility}]))

        # Update layout to add dropdown menu with static title
        main_fig.update_layout(
            title="SHAP Summary Plot: Feature Impact",
            updatemenus=[
                {
                    "buttons": buttons,
                    "direction": "down",
                    "showactive": True,
                    "x": 1.0,
                    "y": 1.15,
                    "xanchor": "right",
                    "yanchor": "top",
                }
            ],
        )
        return main_fig

    def register_internal_callbacks(self):
        """Register internal callbacks for the plugin."""
        pass  # Implement if needed


if __name__ == "__main__":
    """Run the Unit Test for the Plugin."""
    from workbench.web_interface.components.plugin_unit_test import PluginUnitTest

    # Run the Unit Test on the Plugin
    # model = CachedModel("abalone-regression")
    model = CachedModel("wine-classification")
    # model = CachedModel("test-classification")
    PluginUnitTest(ShapSummaryPlot, input_data=model, theme="light").run()



"""Show the current Workbench Config"""

from workbench.utils.config_manager import ConfigManager
from workbench.utils.repl_utils import cprint


# Show the current Workbench Config
def main():
    # Create a ConfigManager instance
    cm = ConfigManager()

    # Show the current Workbench Config
    cprint("yellow", f"\n\nWorkbench Config Path: {cm.site_config_path}")
    config = cm.get_all_config()
    for key, value in config.items():
        cprint(["lightpurple", "\t" + key, "lightgreen", value])


if __name__ == "__main__":
    main()



# Model: feature_space_proximity
#
# Description: The feature_space_proximity model computes nearest neighbors for the given feature space
#

# Template Placeholders
TEMPLATE_PARAMS = {
    "id_column": "udm_mol_bat_id",
    "features": ['bcut2d_logplow', 'numradicalelectrons', 'smr_vsa5', 'fr_lactam', 'fr_morpholine', 'fr_aldehyde', 'slogp_vsa1', 'fr_amidine', 'bpol', 'fr_ester', 'fr_azo', 'kappa3', 'peoe_vsa5', 'fr_ketone_topliss', 'vsa_estate9', 'estate_vsa9', 'bcut2d_mrhi', 'fr_ndealkylation1', 'numrotatablebonds', 'minestateindex', 'fr_quatn', 'peoe_vsa3', 'fr_epoxide', 'fr_aniline', 'minpartialcharge', 'fr_nitroso', 'fpdensitymorgan2', 'fr_oxime', 'fr_sulfone', 'smr_vsa1', 'kappa1', 'fr_pyridine', 'numaromaticrings', 'vsa_estate6', 'molmr', 'estate_vsa1', 'fr_dihydropyridine', 'vsa_estate10', 'fr_alkyl_halide', 'chi2n', 'fr_thiocyan', 'fpdensitymorgan1', 'fr_unbrch_alkane', 'slogp_vsa9', 'chi4n', 'fr_nitro_arom', 'fr_al_oh', 'fr_furan', 'fr_c_s', 'peoe_vsa8', 'peoe_vsa14', 'numheteroatoms', 'fr_ndealkylation2', 'maxabspartialcharge', 'vsa_estate2', 'peoe_vsa7', 'apol', 'numhacceptors', 'fr_tetrazole', 'vsa_estate1', 'peoe_vsa9', 'naromatom', 'bcut2d_chghi', 'fr_sh', 'fr_halogen', 'slogp_vsa4', 'fr_benzodiazepine', 'molwt', 'fr_isocyan', 'fr_prisulfonamd', 'maxabsestateindex', 'minabsestateindex', 'peoe_vsa11', 'slogp_vsa12', 'estate_vsa5', 'numaliphaticcarbocycles', 'bcut2d_mwlow', 'slogp_vsa7', 'fr_allylic_oxid', 'fr_methoxy', 'fr_nh0', 'fr_coo2', 'fr_phenol', 'nacid', 'nbase', 'chi3v', 'fr_ar_nh', 'fr_nitrile', 'fr_imidazole', 'fr_urea', 'bcut2d_mrlow', 'chi1', 'smr_vsa6', 'fr_aryl_methyl', 'narombond', 'fr_alkyl_carbamate', 'fr_piperzine', 'exactmolwt', 'qed', 'chi0n', 'fr_sulfonamd', 'fr_thiazole', 'numvalenceelectrons', 'fr_phos_acid', 'peoe_vsa12', 'fr_nh1', 'fr_hdrzine', 'fr_c_o_nocoo', 'fr_lactone', 'estate_vsa6', 'bcut2d_logphi', 'vsa_estate7', 'peoe_vsa13', 'numsaturatedcarbocycles', 'fr_nitro', 'fr_phenol_noorthohbond', 'rotratio', 'fr_barbitur', 'fr_isothiocyan', 'balabanj', 'fr_arn', 'fr_imine', 'maxpartialcharge', 'fr_sulfide', 'slogp_vsa11', 'fr_hoccn', 'fr_n_o', 'peoe_vsa1', 'slogp_vsa6', 'heavyatommolwt', 'fractioncsp3', 'estate_vsa8', 'peoe_vsa10', 'numaliphaticrings', 'fr_thiophene', 'maxestateindex', 'smr_vsa10', 'labuteasa', 'smr_vsa2', 'fpdensitymorgan3', 'smr_vsa9', 'slogp_vsa10', 'numaromaticheterocycles', 'fr_nh2', 'fr_diazo', 'chi3n', 'fr_ar_coo', 'slogp_vsa5', 'fr_bicyclic', 'fr_amide', 'estate_vsa10', 'fr_guanido', 'chi1n', 'numsaturatedrings', 'fr_piperdine', 'fr_term_acetylene', 'estate_vsa4', 'slogp_vsa3', 'fr_coo', 'fr_ether', 'estate_vsa7', 'bcut2d_chglo', 'fr_oxazole', 'peoe_vsa6', 'hallkieralpha', 'peoe_vsa2', 'chi2v', 'nocount', 'vsa_estate5', 'fr_nhpyrrole', 'fr_al_coo', 'bertzct', 'estate_vsa11', 'minabspartialcharge', 'slogp_vsa8', 'fr_imide', 'kappa2', 'numaliphaticheterocycles', 'numsaturatedheterocycles', 'fr_hdrzone', 'smr_vsa4', 'fr_ar_n', 'nrot', 'smr_vsa8', 'slogp_vsa2', 'chi4v', 'fr_phos_ester', 'fr_para_hydroxylation', 'smr_vsa3', 'nhohcount', 'estate_vsa2', 'mollogp', 'tpsa', 'fr_azide', 'peoe_vsa4', 'numhdonors', 'fr_al_oh_notert', 'fr_c_o', 'chi0', 'fr_nitro_arom_nonortho', 'vsa_estate3', 'fr_benzene', 'fr_ketone', 'vsa_estate8', 'smr_vsa7', 'fr_ar_oh', 'fr_priamide', 'ringcount', 'estate_vsa3', 'numaromaticcarbocycles', 'bcut2d_mwhi', 'chi1v', 'heavyatomcount', 'vsa_estate4', 'chi0v'],
    "target": "udm_asy_res_value",
    "track_columns": None
}

from io import StringIO
import json
import argparse
import os
import pandas as pd

# Local Imports
from proximity import Proximity


# Function to check if dataframe is empty
def check_dataframe(df: pd.DataFrame, df_name: str) -> None:
    """Check if the DataFrame is empty and raise an error if so."""
    if df.empty:
        msg = f"*** The training data {df_name} has 0 rows! ***STOPPING***"
        print(msg)
        raise ValueError(msg)


# Function to match DataFrame columns to model features (case-insensitive)
def match_features_case_insensitive(df: pd.DataFrame, model_features: list) -> pd.DataFrame:
    """Match and rename DataFrame columns to match the model's features, case-insensitively."""
    # Create a set of exact matches from the DataFrame columns
    exact_match_set = set(df.columns)

    # Create a case-insensitive map of DataFrame columns
    column_map = {col.lower(): col for col in df.columns}
    rename_dict = {}

    # Build a dictionary for renaming columns based on case-insensitive matching
    for feature in model_features:
        if feature in exact_match_set:
            rename_dict[feature] = feature
        elif feature.lower() in column_map:
            rename_dict[column_map[feature.lower()]] = feature

    # Rename columns in the DataFrame to match model features
    return df.rename(columns=rename_dict)


# TRAINING SECTION
#
# This section (__main__) is where SageMaker will execute the training job
# and save the model artifacts to the model directory.
#
if __name__ == "__main__":
    # Template Parameters
    id_column = TEMPLATE_PARAMS["id_column"]
    features = TEMPLATE_PARAMS["features"]
    target = TEMPLATE_PARAMS["target"]  # Can be None for unsupervised models
    track_columns = TEMPLATE_PARAMS["track_columns"]  # Can be None

    # Script arguments for input/output directories
    parser = argparse.ArgumentParser()
    parser.add_argument("--model-dir", type=str, default=os.environ.get("SM_MODEL_DIR", "/opt/ml/model"))
    parser.add_argument("--train", type=str, default=os.environ.get("SM_CHANNEL_TRAIN", "/opt/ml/input/data/train"))
    parser.add_argument(
        "--output-data-dir", type=str, default=os.environ.get("SM_OUTPUT_DATA_DIR", "/opt/ml/output/data")
    )
    args = parser.parse_args()

    # Load training data from the specified directory
    training_files = [
        os.path.join(args.train, file)
        for file in os.listdir(args.train) if file.endswith(".csv")
    ]
    all_df = pd.concat([pd.read_csv(file, engine="python") for file in training_files])

    # Check if the DataFrame is empty
    check_dataframe(all_df, "training_df")

    # Create the Proximity model
    model = Proximity(all_df, id_column, features, target, track_columns=track_columns)

    # Now serialize the model
    model.serialize(args.model_dir)

# Model loading and prediction functions
def model_fn(model_dir):

    # Deserialize the model
    model = Proximity.deserialize(model_dir)
    return model


def input_fn(input_data, content_type):
    """Parse input data and return a DataFrame."""
    if not input_data:
        raise ValueError("Empty input data is not supported!")

    # Decode bytes to string if necessary
    if isinstance(input_data, bytes):
        input_data = input_data.decode("utf-8")

    if "text/csv" in content_type:
        return pd.read_csv(StringIO(input_data))
    elif "application/json" in content_type:
        return pd.DataFrame(json.loads(input_data))  # Assumes JSON array of records
    else:
        raise ValueError(f"{content_type} not supported!")


def output_fn(output_df, accept_type):
    """Supports both CSV and JSON output formats."""
    use_explicit_na = False
    if "text/csv" in accept_type:
        if use_explicit_na:
            csv_output = output_df.fillna("N/A").to_csv(index=False)  # CSV with N/A for missing values
        else:
            csv_output = output_df.to_csv(index=False)
        return csv_output, "text/csv"
    elif "application/json" in accept_type:
        return output_df.to_json(orient="records"), "application/json"  # JSON array of records (NaNs -> null)
    else:
        raise RuntimeError(f"{accept_type} accept type is not supported by this script.")


# Prediction function
def predict_fn(df, model):
    # Match column names before prediction if needed
    df = match_features_case_insensitive(df, model.features + [model.id_column])

    # Compute Nearest neighbors
    df = model.neighbors(df)
    return df



from workbench.api.feature_set import FeatureSet
import pandas as pd
from pprint import pprint

# Set Pandas output options
pd.set_option("display.max_columns", None)
pd.set_option("display.max_rows", None)
pd.set_option("display.width", 1000)

# Grab a FeatureSet and pull some of the EDA Stats
my_features = FeatureSet("test_features")

# Grab some of the EDA Stats
corr_data = my_features.correlations()
corr_df = pd.DataFrame(corr_data)

# Sort both the columns and index (for readability)
corr_df.sort_index(inplace=True)
corr_df.sort_index(axis=1, inplace=True)
print(corr_df)

# Get some outliers
outliers = my_features.outliers()
pprint(outliers.head())

# Full set of EDA Stats
eda_stats = my_features.column_stats()
pprint(eda_stats)



import networkx as nx

# Create a graph
G = nx.Graph()

# Add nodes with attributes
G.add_node(1, id=1, name="Alice", age=30)
G.add_node(2, id=2, name="Bob", age=25)

# Print nodes with attributes before serialization
print("Before serialization:")
for node, attrs in G.nodes(data=True):
    print(node, attrs)

# Serialize to JSON
graph_json = nx.readwrite.json_graph.node_link_data(G, edges="edges", name="dummy")
# graph_json = nx.readwrite.json_graph.node_link_data(G, edges="edges")

# Deserialize back
G_deserialized = nx.readwrite.json_graph.node_link_graph(graph_json, edges="edges", name="dummy")
# G_deserialized = nx.readwrite.json_graph.node_link_graph(graph_json, edges="edges")

# Print nodes with attributes after deserialization
print("\nAfter deserialization:")
for node, attrs in G_deserialized.nodes(data=True):
    print(node, attrs)



"""Repl utilities for Workbench"""

import threading
import itertools
import time
import sys

# Colors
colors = {
    "lightblue": "\x1b[38;5;69m",
    "lightpurple": "\x1b[38;5;141m",
    "lightgreen": "\x1b[38;5;113m",
    "lime": "\x1b[38;5;154m",
    "darkyellow": "\x1b[38;5;220m",
    "orange": "\x1b[38;5;208m",
    "red": "\x1b[38;5;198m",
    "pink": "\x1b[38;5;213m",
    "magenta": "\x1b[38;5;206m",
    "tan": "\x1b[38;5;179m",
    "lighttan": "\x1b[38;5;180m",
    "yellow": "\x1b[38;5;226m",
    "green": "\x1b[38;5;34m",
    "blue": "\x1b[38;5;21m",
    "purple": "\x1b[38;5;91m",
    "purple_blue": "\x1b[38;5;63m",
    "lightgrey": "\x1b[38;5;250m",
    "grey": "\x1b[38;5;244m",
    "darkgrey": "\x1b[38;5;240m",
    "reset": "\x1b[0m",
}


def cprint(*args):
    """
    Print text in color. Supports either a single color and text or a list of color-text pairs.

    Args:
         A single color and text or a list of color-text pairs.
         For example: cprint('red', 'Hello') or cprint(['red', 'Hello', 'green', 'World'])
    """
    if isinstance(args[0], list):
        args = args[0]
    # Iterate over the arguments in pairs of color and text
    for i in range(0, len(args), 2):
        print(f"{colors[args[i]]}{args[i + 1]}{colors['reset']}", end=" ")
    print()  # Print a newline at the end


def status_lights(status_colors: list[str]):
    """
    Create status lights (circles) in color.

    Args:
        status_colors: A list of status colors (e.g. ['red', 'green', 'yellow']).

    Returns:
        A string representation of colored status lights.
    """
    circle = "●"  # Unicode character for a filled circle
    return "".join(f"{colors[color]}{circle}{colors['reset']}" for color in status_colors)


# Spinner Class
class Spinner:
    def __init__(self, color, message="Loading..."):
        """
        Initialize the Spinner object.

        Args:
            color: The color of the spinner.
            message: The message to display beside the spinner.
        """
        self.done = False
        self.message = message
        self.color = color
        self.thread = threading.Thread(target=self.spin)

    def _write(self, text):
        """Write text to stdout and flush."""
        sys.stdout.write(text)
        sys.stdout.flush()

    def _hide_cursor(self):
        """Hide the terminal cursor."""
        self._write("\033[?25l")

    def _show_cursor(self):
        """Show the terminal cursor."""
        self._write("\n\033[?25h")

    def spin(self):
        """Display a multi-spinner animation until stopped."""
        frames = "⠋⠙⠹⠸⠼⠴⠦⠧⠇⠏"  # Frames used for each spinner
        done_frame = "⠿⠿⠿⠿"  # Frames to show on completion

        # Create four separate cycle iterators for the spinners
        spinners = [itertools.cycle(frames) for _ in range(4)]

        # Initialize each spinner with staggered positions for visual effect
        for i, spinner in enumerate(spinners):
            for _ in range(i * 2):  # Stagger each spinner by 2 frames
                next(spinner)

        self._hide_cursor()
        while not self.done:
            # Construct the spinner display from each staggered spinner
            spinner_display = "".join(next(spinner) for spinner in spinners)
            self._write(f"\r{colors[self.color]}{self.message} {colors['darkyellow']}{spinner_display}")
            time.sleep(0.1)  # Control the speed of spinner animation

        # Display the "done" frame when completed
        self._write(f"\r{colors[self.color]}{self.message} {colors['lightgreen']}{done_frame}")
        self._show_cursor()

    def start(self):
        """Start the spinner in a separate thread."""
        self.thread.start()

    def stop(self):
        """Stop the spinner and wait for the thread to finish."""
        self.done = True
        self.thread.join()  # Ensure the spinner thread is fully stopped


if __name__ == "__main__":
    # Print all the colors for a quick visual test
    for color, _ in colors.items():
        cprint(color, f"Hello world! ({color})")

    # Print a list of color-text pairs
    cprint(["red", "Hello", "green", "World"])

    # Print a string of status lights
    print(status_lights(["red", "green", "yellow"]))

    # Spinner Testing
    def long_running_operation():
        """Simulate a long-running operation."""
        time.sleep(3)

    # Test the spinner with a sample message
    spinner = Spinner("lightpurple", "Chatting with AWS:")
    spinner.start()
    try:
        long_running_operation()  # Simulate a long-running task
    finally:
        spinner.stop()
        print("Done!")



import os

import aws_cdk as cdk

# TODO Make sure this import works in new folder structure
from workbench_image_update_stack import WorkbenchImageUpdateStack


app = cdk.App()
WorkbenchImageUpdateStack(
    app,
    "WorkbenchImageUpdate",
    # If you don't specify 'env', this stack will be environment-agnostic.
    # Account/Region-dependent features and context lookups will not work,
    # but a single synthesized template can be deployed anywhere.
    # Uncomment the next line to specialize this stack for the AWS Account
    # and Region that are implied by the current CLI configuration.
    # env=cdk.Environment(account=os.getenv('CDK_DEFAULT_ACCOUNT'), region=os.getenv('CDK_DEFAULT_REGION')),
    # Uncomment the next line if you know exactly what Account and Region you
    # want to deploy the stack to. */
    # env=cdk.Environment(account='123456789012', region='us-east-1'),
    # For more information, see https://docs.aws.amazon.com/cdk/latest/guide/environments.html
)

app.synth()



import pandas as pd
import numpy as np
from sklearn.neighbors import NearestNeighbors
from typing import Union, List
import logging

# Workbench Imports
from workbench.algorithms.dataframe.proximity import Proximity, ProximityType

# Set up logging
log = logging.getLogger("workbench")


class FingerprintProximity(Proximity):
    def __init__(
        self, df: pd.DataFrame, id_column: Union[int, str], fingerprint_column: str, n_neighbors: int = 5
    ) -> None:
        """
        Initialize the FingerprintProximity class for binary fingerprint similarity.

        Args:
            df (pd.DataFrame): DataFrame containing fingerprints.
            id_column (Union[int, str]): Name of the column used as an identifier.
            fingerprint_column (str): Name of the column containing fingerprints.
            n_neighbors (int): Default number of neighbors to compute.
        """
        self.fingerprint_column = fingerprint_column

        # Call the parent class constructor
        super().__init__(df, id_column=id_column, features=[fingerprint_column], n_neighbors=n_neighbors)

    # Override the build_proximity_model method
    def build_proximity_model(self) -> None:
        """
        Prepare the fingerprint data for nearest neighbor calculations.
        Converts fingerprint strings to binary arrays and initializes NearestNeighbors.
        """
        log.info("Converting fingerprints to binary feature matrix...")
        self.proximity_type = ProximityType.SIMILARITY

        # Convert fingerprint strings to binary arrays

        fingerprint_bits = self.df[self.fingerprint_column].apply(
            lambda fp: np.array([int(bit) for bit in fp], dtype=np.bool_)
        )
        self.X = np.vstack(fingerprint_bits)

        # Use Jaccard similarity for binary fingerprints
        log.info("Computing NearestNeighbors with Jaccard metric...")
        self.nn = NearestNeighbors(metric="jaccard", n_neighbors=self.n_neighbors + 1).fit(self.X)

    # Override the prep_features_for_query method
    def prep_features_for_query(self, query_df: pd.DataFrame) -> np.ndarray:
        """
        Prepare the query DataFrame by converting fingerprints to binary arrays.

        Args:
            query_df (pd.DataFrame): DataFrame containing query fingerprints.

        Returns:
            np.ndarray: Binary feature matrix for the query fingerprints.
        """
        fingerprint_bits = query_df[self.fingerprint_column].apply(
            lambda fp: np.array([int(bit) for bit in fp], dtype=np.bool_)
        )
        return np.vstack(fingerprint_bits)

    def all_neighbors(
        self,
        min_similarity: float = None,
        include_self: bool = False,
        add_columns: List[str] = None,
    ) -> pd.DataFrame:
        """
        Find neighbors for all fingerprints in the dataset.

        Args:
            min_similarity: Minimum similarity threshold (0-1)
            include_self: Whether to include self in results
            add_columns: Additional columns to include in results

        Returns:
            DataFrame containing neighbors and similarities
        """

        # Call the parent class method to find neighbors
        return self.neighbors(
            query_df=self.df,
            min_similarity=min_similarity,
            include_self=include_self,
            add_columns=add_columns,
        )

    def neighbors(
        self,
        query_df: pd.DataFrame,
        min_similarity: float = None,
        include_self: bool = False,
        add_columns: List[str] = None,
    ) -> pd.DataFrame:
        """
        Find neighbors for each row in the query DataFrame.

        Args:
            query_df: DataFrame containing query fingerprints
            min_similarity: Minimum similarity threshold (0-1)
            include_self: Whether to include self in results (if present)
            add_columns: Additional columns to include in results

        Returns:
            DataFrame containing neighbors and similarities

        Note: The query DataFrame must include the feature columns. The id_column is optional.
        """

        # Calculate radius from similarity if provided
        radius = 1 - min_similarity if min_similarity is not None else None

        # Call the parent class method to find neighbors
        neighbors_df = super().neighbors(
            query_df=query_df,
            radius=radius,
            include_self=include_self,
            add_columns=add_columns,
        )

        # Convert distances to similarity
        neighbors_df["similarity"] = 1 - neighbors_df["distance"]
        neighbors_df.drop(columns=["distance"], inplace=True)
        return neighbors_df


# Testing the FingerprintProximity class
if __name__ == "__main__":
    pd.set_option("display.max_columns", None)
    pd.set_option("display.width", 1000)

    # Example DataFrame
    data = {
        "id": ["a", "b", "c", "d"],
        "fingerprint": ["101010", "111010", "101110", "011100"],
    }
    df = pd.DataFrame(data)

    # Initialize the FingerprintProximity class
    proximity = FingerprintProximity(df, fingerprint_column="fingerprint", id_column="id", n_neighbors=3)

    # Test 1: All neighbors
    print("\n--- Test 1: All Neighbors ---")
    all_neighbors_df = proximity.all_neighbors()
    print(all_neighbors_df)

    # Test 2: Neighbors for a specific query
    print("\n--- Test 2: Neighbors for Query ---")
    query_df = pd.DataFrame({"id": ["a"], "fingerprint": ["101010"]})
    query_neighbors_df = proximity.neighbors(query_df=query_df)
    print(query_neighbors_df)

    # Test 3: Neighbors with similarity threshold
    print("\n--- Test 3: Neighbors with Minimum Similarity 0.5 ---")
    query_neighbors_sim_df = proximity.neighbors(query_df=query_df, min_similarity=0.5)
    print(query_neighbors_sim_df)



import argparse
import numpy as np
import pandas as pd

#
from workbench.api import DataSource
from workbench.core.artifacts.athena_source import AthenaSource
from workbench.core.transforms.pandas_transforms import PandasToData
from workbench.core.cloud_platform.aws.aws_meta import AWSMeta


def compare_dataframes_with_tolerance(df1: pd.DataFrame, df2: pd.DataFrame, tolerance: float = 1e-8) -> bool:
    # Check column names
    if list(df1.columns) != list(df2.columns):
        print("Column names differ:")
        print(f"df1 columns: {list(df1.columns)}")
        print(f"df2 columns: {list(df2.columns)}")
        return False

    # Check column types
    if not all(df1.dtypes == df2.dtypes):
        print("Column types differ:")
        print("df1 types:")
        print(df1.dtypes)
        print("df2 types:")
        print(df2.dtypes)
        return False

    # Check number of rows
    if len(df1) != len(df2):
        print(f"Number of rows differ: df1 has {len(df1)} rows, df2 has {len(df2)} rows.")
        return False

    # Sort dataframes by ID columns
    id_columns = ["id", "ID", "primary_key", "unique_id"]
    sort_columns = [col for col in id_columns if col in df1.columns and col in df2.columns]

    if sort_columns:
        df1 = df1.sort_values(by=sort_columns).reset_index(drop=True)
        df2 = df2.sort_values(by=sort_columns).reset_index(drop=True)

    # Compare each column
    for column in df1.columns:
        if df1[column].dtype.kind in "f":  # Float column
            is_close = np.isclose(df1[column], df2[column], atol=tolerance)
            if not np.all(is_close):
                print(f"Differences found in column: {column}")
                return False
        else:  # Non-float column
            if not df1[column].equals(df2[column]):
                print(f"Differences found in column: {column}")
                return False

    # DataFrames are equivalent
    return True


def migrate(ds_name):
    old_database = "sageworks"

    # Get the old datasource and grab a full dataframe
    print(f"Fetching data from old DataSource: {ds_name} in database: {old_database}")
    old_ds = AthenaSource(ds_name, database=old_database)
    old_df = old_ds.query(f"SELECT * FROM {old_ds.table}")

    # Create the new DataSource (defaults to workbench database)
    print(f"Migrating DataSource {ds_name} to the new database...")
    df_to_data = PandasToData(ds_name)  # Use the same name for the new DataSource
    df_to_data.set_output_tags(old_ds.get_tags())  # Transfer tags from the old DataSource
    df_to_data.set_input(old_df)  # Pass the data from the old DataSource
    df_to_data.transform()  # Transform and save the new DataSource

    print(f"Validating the new DataSource {ds_name}...")
    new_ds = DataSource(ds_name)
    new_df = new_ds.query(f"SELECT * FROM {new_ds.table}")

    # Compare the old and new DataFrames
    if compare_dataframes_with_tolerance(old_df, new_df, tolerance=1e-8):
        print(f"Migration Successful: DataSource {ds_name} has been migrated to the new database.")
    else:
        print(f"Migration Failed: DataSource {ds_name} dataframe did not match the original.")


def migrate_all():
    old_database = "sageworks"

    # Grab the DataSource in the old database
    old_datasources = AWSMeta()._list_catalog_tables(old_database)
    for ds_name in old_datasources["Name"].values:
        print(f"Starting migration for DataSource: {ds_name}")
        try:
            # print(f"Mock migration for {ds_name}")
            migrate(ds_name)
        except Exception as e:
            print(f"Migration failed for: {ds_name}. Error: {e}")


if __name__ == "__main__":
    # Argument parsing to choose between single or all DataSource migration
    parser = argparse.ArgumentParser(description="Migrate DataSources from the old database to the new one.")
    parser.add_argument(
        "--all",
        action="store_true",
        help="Migrate all DataSources from the old database",
    )
    parser.add_argument(
        "--ds_name",
        type=str,
        help="The name of a single DataSource to migrate (optional if --all is used)",
    )
    args = parser.parse_args()

    if args.all:
        migrate_all()
    elif args.ds_name:
        migrate(args.ds_name)
    else:
        print("Please specify --all to migrate all DataSources or --ds_name for a single DataSource.")



"""SQL based Correlations: Compute Correlations for all the numeric columns in a DataSource using SQL"""

import logging
import pandas as pd
from collections import defaultdict

# Workbench Imports
from workbench.core.artifacts.data_source_abstract import DataSourceAbstract


# Workbench Logger
log = logging.getLogger("workbench")


def correlation_query(columns: list[str], table_name: str) -> str:
    """Build a query to compute the correlations between columns in a table

    Args:
        columns (list(str)): The columns to compute correlations on
        table_name (str): The table to compute correlations on

    Returns:
        str: The SQL query to compute correlations

    Notes: Pearson correlation coefficient ranges from -1 to 1:
           +1 indicates a perfect positive linear relationship.
           -1 indicates a perfect negative linear relationship.
            0 indicates no linear relationship.
    """
    query = f'SELECT <<cross_correlations>> FROM "{table_name}"'
    cross_correlations = ""
    for c in columns:
        for d in columns:
            if c != d:
                cross_correlations += f'corr("{c}", "{d}") AS "{c}__{d}", '
    query = query.replace("<<cross_correlations>>", cross_correlations[:-2])

    # Return the query
    return query


def correlations(data_source: DataSourceAbstract) -> dict[dict]:
    """Compute Correlations for all the numeric columns in a DataSource
    Args:
        data_source(DataSource): The DataSource that we're computing correlations on
    Returns:
        dict(dict): A dictionary of correlations for each column in this format
             {'col1': {'col2': 0.5, 'col3': 0.9, 'col4': 0.4, ...},
              'col2': {'col1': 0.5, 'col3': 0.8, 'col4': 0.3, ...}}
    """
    data_source.log.info("Computing Correlations for numeric columns...")

    # Figure out which columns are numeric
    num_type = ["double", "float", "int", "bigint", "smallint", "tinyint"]
    details = data_source.view("computation").column_details()

    # Get the numeric columns
    numeric = [column for column, data_type in details.items() if data_type in num_type]

    # If we have at least two numeric columns, compute the correlations
    if len(numeric) < 2:
        return {}

    # Grab the DataSource computation table name
    table = data_source.view("computation").table

    # Build the query
    query = correlation_query(numeric, table)

    # Run the query
    log.debug(query)
    result_df = data_source.query(query)

    # Drop any columns that have NaNs
    result_df = result_df.dropna(axis=1)

    # Process the results
    # Note: The result_df is a DataFrame with a single row and a column for each pairwise correlation
    correlation_dict = result_df.to_dict(orient="index")[0]

    # Convert the dictionary to a nested dictionary
    # Note: The keys are in the format col1__col2
    nested_corr = defaultdict(dict)
    for key, value in correlation_dict.items():
        col1, col2 = key.split("__")
        nested_corr[col1][col2] = value

    # Sort the nested dictionaries
    sorted_dict = {}
    for key, sub_dict in nested_corr.items():
        sorted_dict[key] = {k: v for k, v in sorted(sub_dict.items(), key=lambda item: item[1], reverse=True)}
    return sorted_dict


if __name__ == "__main__":
    """Exercise the SQL Correlations Functionality"""
    from workbench.api.data_source import DataSource

    # Setup Pandas output options
    pd.set_option("display.max_colwidth", 50)
    pd.set_option("display.max_columns", 15)
    pd.set_option("display.width", 1000)

    # Retrieve a Data Source
    my_data = DataSource("test_data")

    # Verify that the Athena Data Source exists
    assert my_data.exists()

    # What's my Workbench Name
    print(f"Name: {my_data.name}")

    # Get correlations for numeric columns
    my_correlations = correlations(my_data)
    print("\nCorrelations")
    print(my_correlations)
